# Copyright (c) 2022-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Helper functions for LoadContext operator."""

import re

import numpy as np
from astropy.table import Column

import so3g

import toast
from toast.mpi import MPI
from toast.timing import function_timer, Timer
from toast.utils import Logger
from toast.dist import distribute_uniform
from toast.observation import default_values as defaults

from ...coords import pointing_model
from ...core import Context, AxisManager
from ...core.axisman import AxisInterface
from ...preprocess import Pipeline as PreProcPipe
from ...hwp.hwp_angle_model import apply_hwp_angle_model


def open_context(context=None, context_file=None):
    if context is None:
        # The user did not specify a context- create a temporary
        # one from the file
        return Context(context_file)
    else:
        # Just return the user-specified context.
        return context


@function_timer
def read_and_preprocess_wafers(
    obs_name,
    session_name,
    gcomm,
    wafer_readers,
    wafer_dets,
    preconfig=None,
    ignore_preprocess_archive=False,
    context=None,
    context_file=None,
):
    """Read the wafer data.

    Each process reads zero or more wafers and applies preprocessing.

    If the preprocessing config is specified, it should include the archive stanza,
    which will be inserted into the context dictionary.

    Args:
        obs_name (str):  The observation name.
        session_name (str):  The observing session.
        gcomm (MPIComm):  The observation group communicator, or None.
        wafer_readers (dict):  For each wafer name, the group rank assigned
            to read this wafer.
        wafer_dets (dict):  For each wafer name, the list of detectors.
        preconfig (dict):  The preprocessing configuration to apply (or None).
        ignore_preprocess_db (bool): Ignore the 'archive' field in the preprocessing configuration.
        context (Context):  The pre-existing Context or None.
        context_file (str):  The context file to open or None.

    Returns:
        (dict):  The AxisManager data for each wafer on this process.

    """
    log = Logger.get()
    timer = Timer()
    timer.start()
    if gcomm is None:
        rank = 0
    else:
        rank = gcomm.rank

    results = dict()
    for wf, reader in wafer_readers.items():
        if reader == rank:
            ctx = open_context(context=context, context_file=context_file)
            if (preconfig is not None) and (not ignore_preprocess_archive):
                # Ensure that the preprocessing archive is defined.  First
                # check if it already exists in the context.
                have_ctx_archive = False
                if ctx.get("metadata") is None:
                    ctx["metadata"] = []
                for key in ctx.get("metadata"):
                    if key.get("name") == "preprocess":
                        have_ctx_archive = True
                        break
                # Check if the archive exists in the preprocess config
                if "archive" in preconfig:
                    have_pre_archive = True
                else:
                    have_pre_archive = False
                # The archive should only be defined in one place...
                if have_ctx_archive and have_pre_archive:
                    msg = "Both the context file AND the preprocess config define"
                    msg += " a preprocess archive.  There can be only one."
                    raise RuntimeError(msg)
                if not have_ctx_archive:
                    if not have_pre_archive:
                        msg = "Either the context or the preprocess config must "
                        msg += "specify the preprocess archive. "
                        msg += "If you want to ignore preprocessing archive, "
                        msg += "please set ignore_preprocess_archive=True."
                        raise RuntimeError(msg)
                    else:
                        ctx["metadata"].append(
                            {"db": preconfig["archive"]["index"], "name": "preprocess"}
                        )
            # Load the data and immediately close context
            axtod = ctx.get_obs(session_name, dets=wafer_dets[wf])
            if context_file is not None:
                del ctx
            timer.stop()
            elapsed = timer.seconds()
            timer.start()
            log.debug(
                f"LoadContext {obs_name} loaded wafer {wf} in {elapsed} seconds",
            )

            # Apply pointing model, UNLESS we are applying a preprocess config
            # on load and that config includes the pointing model
            if "pointing_model" in axtod:
                if (
                    preconfig is None
                    or "pointing_model" not in [i['name'] for i in preconfig["process_pipe"]]
                ):
                    pointing_model.apply_pointing_model(axtod)

            # If the axis manager has a HWP angle solution, apply it.
            if "hwp_solution" in axtod:
                # Did we already apply it in the preprocessing?
                if (
                    preconfig is None
                    or "hwp_angle_model" not in [i['name'] for i in preconfig["process_pipe"]]
                ):
                    axtod = apply_hwp_angle_model(axtod)
                    timer.stop()
                    elapsed = timer.seconds()
                    timer.start()
                    log.debug(
                        f"LoadContext {obs_name} HWP model wafer {wf} in {elapsed} seconds",
                    )

            if preconfig is not None:
                prepipe = PreProcPipe(preconfig["process_pipe"], logger=log)
                if not ignore_preprocess_archive:
                    prepipe.run(axtod, axtod.preprocess)
                else:
                    prepipe.run(axtod)
                timer.stop()
                elapsed = timer.seconds()
                timer.start()
                msg = f"LoadContext {obs_name} apply preproc to {wf}"
                msg += f" in {elapsed} seconds"
                log.debug(msg)
            results[wf] = axtod
    return results


def parse_metadata(axman, obs_meta, fp_cols, path_sep, det_axis, obs_base, fp_base):
    """Parse Context metadata.

    This is a recursive function which starts with the AxisManger returned
    by `Context.get_meta()` and splits meta data into "detector properties"
    which will be placed in the focalplane model and "other metadata" which
    will be stored in the Observation dictionary.

    For each element of the input AxisManager we do the following:

    - If the object has one axis and it is the detector axis,
      treat it as a column in the detector property table
    - If the object has one axis and it is the sample axis,
      it will be loaded later as shared data.
    - If the object has multiple axes, load it later
    - If the object has no axes, then treat it as observation
      metadata
    - If the object is a nested AxisManager, descend and apply
      the same steps as above.

    Args:
        axman (AxisManager):  The input axismanager.
        obs_meta (dict):  The working dictionary of observation metadata.
        fp_cols (dict):  Dictionary of table Columns of detector properties.
        path_sep (str):  The path separation char when building flattened names.
        det_axis (str):  The name of the detector axis.
        obs_base (str):  The current key in the nested observation metadata.
        fp_base (str):  The current key to use when appending focalplane columns.

    Returns:
        None

    """
    if obs_base is None:
        om = obs_meta
    else:
        obs_meta[obs_base] = dict()
        om = obs_meta[obs_base]
    for key in axman.keys():
        if fp_base is not None:
            fp_key = f"{fp_base}{path_sep}{key}"
        else:
            fp_key = key
        if isinstance(axman[key], AxisInterface):
            # This is one of the axes
            continue
        if isinstance(axman[key], AxisManager):
            # Descend
            parse_metadata(
                axman[key], obs_meta, fp_cols, path_sep, det_axis, key, fp_key
            )
        else:
            # FIXME:  It would be nicer if this information was available
            # through a public member...
            field_axes = axman._assignments[key]
            if len(field_axes) == 0:
                # This data is not associated with an axis.
                om[key] = axman[key]
            elif len(field_axes) == 1 and field_axes[0] == det_axis:
                # This is a detector property
                if fp_key in fp_cols:
                    msg = f"Context meta key '{fp_key}' is duplicated in nested"
                    msg += " AxisManagers"
                    raise RuntimeError(msg)
                fp_cols[fp_key] = Column(name=fp_key, data=np.array(axman[key]))
    if obs_base is not None and len(om) == 0:
        # There were no meta data keys- delete this dict
        del obs_meta[obs_base]


def _local_process_dets(obs, rank):
    """Get the list of local detectors for a specific rank."""
    # Full detector list
    det_names = obs.all_detectors
    # The range of full detector indices on this process.
    full_indices = obs.dist.det_indices[rank]
    # The same info, as a slice.
    full_slc = slice(full_indices.offset, full_indices.offset + full_indices.n_elem, 1)
    # The names of the detectors on this process.
    return det_names[full_slc]


def distribute_detector_props(obs, wafer_key):
    """Compute the communication patterns for detector data.

    This computes the range of local detector indices for each wafer on all
    processes, and also the corresponding range of indices in the wafer data
    on the reading process.  For example, imagine we have 2 wafers and 4
    processes:

    +====================+    +====================+
    |       Wafer 0      |    |     Process 0      |
    |                    |    |--------------------|
    |                    |    |     Process 1      |
    +====================+    |                    |
    |       Wafer 1      |    |--------------------|
    |                    |    |     Process 2      |
    |                    |    |--------------------|
    |                    |    |     Process 3      |
    +====================+    +--------------------+

    In this case, Process 0 will read Wafer 0 and Process 1 will read Wafer 1.
    The data from Wafer 0 needs to be sent to Processes 0 and 1, and the data
    from Wafer 1 needs to be sent to Processes 1, 2, and 3.

    Args:
        obs (Observation):  The observation.
        wafer_key (str):  The column of the focalplane table with the wafer name.

    Returns:
        (tuple):  The (wafer_dets, wafer_readers, wafer_proc_dets, proc_wafer_dets)

    """
    gcomm = obs.comm.comm_group
    rank = obs.comm.group_rank
    gsize = obs.comm.group_size

    det_table = obs.telescope.focalplane.detector_data

    # Wafer names for all detectors
    det_wafers = list(det_table[wafer_key])

    # For each process, this has a key for each wafer on that process.
    # The value for each wafer key is the destination local detector indices for
    # that wafer.
    proc_wafer_dets = dict()

    # For each wafer, this has a key for each destination process.  The value
    # for each process key is the range of detector indices in the wafer data
    # that should be sent to that process.
    wafer_proc_dets = dict()

    # For each wafer, the process that should do the reading.  This is the first
    # process rank that has any data from that wafer.
    wafer_readers = dict()

    # For each wafer, the list of local detector names within that wafer.
    wafer_dets = dict()

    wafer_off = 0
    cur_wafer = det_wafers[0]
    wafer_proc_dets[cur_wafer] = dict()
    wafer_readers[cur_wafer] = 0
    wafer_dets[cur_wafer] = list()
    for proc in range(gsize):
        proc_wafer_dets[proc] = dict()
        # The range of full detector indices on this process.
        full_indices = obs.dist.det_indices[proc]
        # The same info, as a slice.
        full_slc = slice(
            full_indices.offset, full_indices.offset + full_indices.n_elem, 1
        )
        # The names of the detectors on this process.
        local_det_names = _local_process_dets(obs, proc)
        # The wafer names of the detectors on this process.
        local_det_wafers = det_wafers[full_slc]

        if local_det_wafers[0] != cur_wafer:
            # This process starts on a new wafer
            cur_wafer = local_det_wafers[0]
            wafer_off = 0
        # Check for wafer transitions in this process's data
        wafer_change = [
            x
            for x, (y, z) in enumerate(zip(local_det_wafers[:-1], local_det_wafers[1:]))
            if y != z
        ]
        # Add the detector ranges to the mapping dictionaries
        wfirst = 0
        for wc in wafer_change:
            wlast = wc
            nwblock = wlast - wfirst + 1
            # Extend list of detectors for this wafer
            wafer_dets[cur_wafer].extend(local_det_names[wfirst : wlast + 1])
            # Detector indices for this wafer in the local data
            proc_wafer_dets[proc][cur_wafer] = (wfirst, wlast + 1)
            # Detector indices for this process's local data within the full
            # wafer data.
            wafer_proc_dets[cur_wafer][proc] = (wafer_off, wafer_off + nwblock)
            wfirst = wlast + 1
            # We are now on a new wafer.  Reset the wafer detector offset and
            # also assign the reading to this process.
            cur_wafer = local_det_wafers[wlast + 1]
            wafer_off = 0
            wafer_dets[cur_wafer] = list()
            wafer_readers[cur_wafer] = proc
            wafer_proc_dets[cur_wafer] = dict()
        # Handle final block
        wlast = len(local_det_wafers) - 1
        nwblock = wlast - wfirst + 1
        wafer_dets[cur_wafer].extend(local_det_names[wfirst : wlast + 1])
        proc_wafer_dets[proc][cur_wafer] = (wfirst, wlast + 1)
        wafer_proc_dets[cur_wafer][proc] = (wafer_off, wafer_off + nwblock)
        wafer_off += nwblock

    return (wafer_dets, wafer_readers, wafer_proc_dets, proc_wafer_dets)


@function_timer
def distribute_detector_data(
    obs,
    field,
    axwafers,
    axis_dets,
    axis_samples,
    axfield,
    axdtype,
    wafer_readers,
    wafer_proc_dets,
    proc_wafer_dets,
    path_sep,
    is_flag=False,
    flag_invert=False,
    flag_mask=None,
):
    """Communicate detector data from the reading processes to the destinations.

    Data for each wafer is loaded on exactly one process.  This function uses the
    pre-computed data communication pattern built by `distribute_detector_props()`.
    However, the loaded wafer data AxisManager may have been truncated with a
    symmetric sample buffer on either side and some detectors may have been removed.
    This function must account for this when copying data from the AxisManager into
    The full-size observation.

    The observation data is modified in-place.

    Args:
        obs (Observation):  The observation.
        field (str):  The detdata field in the Observation.
        axwafers (dict):  The dictionary of wafer data loaded on this process.
        axis_dets (str):  The name of the detector LabelAxis in the AxisManagers.
        axis_samples (str):  The name of the sample OffsetAxis in the AxisManagers.
        axfield (str):  The name of the data field in the AxisManagers.
        axdtype (np.dtype):  The dtype of the source buffer for sending the data
            from the AxisManager.
        wafer_readers (dict):  The reading process for each wafer.
        wafer_proc_dets (dict):  For each wafer, the dictionary of destination
            processes and the detector indices to send.
        proc_wafer_dets (dict):  For each process, the dictionary of wafer name to
            local detector indices to receive.
        path_sep (str):  The path separation char when building flattened names.
        is_flag (bool):  If True, this field is a flag.
        flag_invert (bool):  If True, invert the meaning of the flag values.
        flag_mask (np.uint8):  The flag mask (or None).

    Returns:
        None

    """
    log = Logger.get()
    gcomm = obs.comm.comm_group
    rank = obs.comm.group_rank
    gsize = obs.comm.group_size

    # If the blocks of detector data exceed 2^30 elements in total, they might hit
    # MPI limitations on the communication message sizes.  Work around that here.
    try:
        from mpi4py.util import pkl5

        if gcomm is not None:
            gcomm = pkl5.Intracomm(gcomm)
    except Exception:
        pass

    def _process_flag(mask, inbuf, detdata, det_begin, det_end, invert):
        n_det = det_end - det_begin
        temp = mask * np.ones_like(inbuf)
        if invert:
            temp[inbuf != 0] = 0
        else:
            temp[inbuf == 0] = 0
        for idet in range(n_det):
            detdata[idet + det_begin, :] |= temp[idet]

    # Keep a handle to our send buffers to ensure that any temporary objects
    # remain in existance until after all receives have happened.
    send_data = dict()
    send_req = list()
    recv_data = None

    # The tag value stride, to ensure unique tags for every sender / receiver / wafer
    # combination.
    tag_stride = 2000

    wf_index = {y: x for x, y in enumerate(wafer_readers.keys())}

    for wafer, reader in wafer_readers.items():
        if reader == rank:
            send_data[wafer] = dict()
            # The wafer axis manager may have a different set of detectors than
            # the full data, if some detectors have been cut.  It may also have
            # a symmetric buffer of samples cut on either side of the observation.
            # here we construct the mapping from full data into this restricted
            # axis manager.
            restricted_dets = axwafers[wafer][axis_dets].vals
            restricted_indices = {y: x for x, y in enumerate(restricted_dets)}

            # If the field is nested, descend
            field_path = axfield.split(path_sep)
            axroot = axwafers[wafer]
            axobject = axroot[field_path[0]]
            for ch in field_path[1:]:
                axobject = axobject[ch]

            for receiver, send_dets in wafer_proc_dets[wafer].items():
                # "send_dets" is the un-restricted range of wafer detectors.
                # We will send a full-size (unrestricted) buffer to ease
                # bookkeeping, and just copy un-cut detector data into that
                # buffer.
                n_send_det = send_dets[1] - send_dets[0]

                # The names of the detectors on the receiving process.
                local_det_names = _local_process_dets(obs, receiver)

                # The receiving detector names and their relative indices
                recv_dets = proc_wafer_dets[receiver][wafer]
                n_recv_det = recv_dets[1] - recv_dets[0]
                recv_det_names = local_det_names[recv_dets[0] : recv_dets[1]]
                recv_det_indices = {y: x for x, y in enumerate(recv_det_names)}

                # Build the mapping of restricted indices to send buffer indices
                restrict_to_send = {
                    restricted_indices[x]: y
                    for x, y in recv_det_indices.items()
                    if x in restricted_indices
                }

                # The per-detector flags, so that detectors are cut properly on the
                # receiver.
                det_flags = np.ones(
                    n_send_det,
                    dtype=np.uint8,
                )
                for idet_ax, idet_send in restrict_to_send.items():
                    det_flags[idet_send] = 0

                # Does the axis manager have a truncated number of samples?
                restricted_samps = axwafers[wafer][axis_samples].count
                if restricted_samps != obs.n_local_samples:
                    ax_shift = axwafers[wafer][axis_samples].offset
                else:
                    ax_shift = 0

                # flat-packed buffer size of send buffer
                flat_size = n_send_det * obs.n_local_samples

                # Is this some detector flag data using ranges instead of samples?
                # If so, we construct a temporary buffer and build sample flags
                # from the ranges.
                if isinstance(axobject, so3g.proj.RangesMatrix):
                    # Yes, flagged ranges.  We may have restricted sample ranges
                    # for our flags, and so we initialize the full buffer to
                    # the invalid mask.
                    sdata = defaults.det_mask_invalid * np.ones(
                        flat_size,
                        dtype=np.uint8,
                    )
                    if flag_invert:
                        for idet_ax, idet_send in restrict_to_send.items():
                            # Set this detector's values to the mask, and then
                            # we will "unflag" the specified ranges.
                            off = idet_send * obs.n_local_samples + ax_shift
                            sdata[off : off + restricted_samps] = flag_mask
                            for rg in axobject[idet_ax].ranges():
                                sdata[off + rg[0] : off + rg[1]] = 0
                    else:
                        for idet_ax, idet_send in restrict_to_send.items():
                            # Set this detector's values to good, and then
                            # we will flag the specified ranges.
                            off = idet_send * obs.n_local_samples + ax_shift
                            sdata[off : off + restricted_samps] = 0
                            for rg in axobject[idet_ax].ranges():
                                sdata[off + rg[0] : off + rg[1]] = flag_mask
                else:
                    # Either normal sample flags or signal data
                    if is_flag:
                        # We may have restricted sample ranges for our flags, and so
                        # we initialize the full buffer to the invalid mask.
                        sdata = defaults.det_mask_invalid * np.ones(
                            flat_size,
                            dtype=axdtype,
                        )
                    else:
                        # Signal data is initialized to zero.  Restricted / missing
                        # samples will be indicated by flags.
                        sdata = np.zeros(
                            flat_size,
                            dtype=axdtype,
                        )
                    for idet_ax, idet_send in restrict_to_send.items():
                        off = idet_send * obs.n_local_samples + ax_shift
                        sdata[off : off + restricted_samps] = axobject[idet_ax, :]

                if receiver == rank:
                    # We just need to process the data locally
                    sdata_2d = sdata.reshape((n_send_det, -1))
                    if is_flag:
                        _process_flag(
                            flag_mask,
                            sdata_2d,
                            obs.detdata[field],
                            recv_dets[0],
                            recv_dets[1],
                            flag_invert,
                        )
                    else:
                        for idet in range(n_send_det):
                            obs.detdata[field][idet + recv_dets[0], :] = sdata_2d[idet]
                    # Update per-detector flags
                    dflags = {
                        obs.local_detectors[recv_dets[0] + x]: defaults.det_mask_invalid
                        for x in range(n_recv_det)
                        if det_flags[x] != 0
                    }
                    obs.update_local_detector_flags(dflags)
                else:
                    # Send asynchronously.
                    tag = (rank * gsize + receiver) * tag_stride + 2 * wf_index[wafer]
                    flag_tag = tag + 1
                    req = gcomm.isend(sdata, dest=receiver, tag=tag)
                    req_flags = gcomm.isend(det_flags, dest=receiver, tag=flag_tag)
                    # Save a handle to this buffer while send operation is in progress
                    send_data[wafer][axfield] = sdata
                    send_data[wafer][f"{axfield}_flags"] = det_flags
                    send_req.append(req)
                    send_req.append(req_flags)

    my_wafer_dets = proc_wafer_dets[rank]
    for wafer, recv_dets in my_wafer_dets.items():
        n_recv_det = recv_dets[1] - recv_dets[0]
        sender = wafer_readers[wafer]
        if sender == rank:
            # This data was already copied locally above
            continue
        else:
            # Receive from sender.  We always allocate a contiguous temporary buffer.
            tag = (sender * gsize + rank) * tag_stride + 2 * wf_index[wafer]
            flag_tag = tag + 1
            recv_data = gcomm.recv(source=sender, tag=tag)
            recv_2d = recv_data.reshape((n_recv_det, -1))
            det_flags = gcomm.recv(source=sender, tag=flag_tag)
            if is_flag:
                _process_flag(
                    flag_mask,
                    recv_2d,
                    obs.detdata[field],
                    recv_dets[0],
                    recv_dets[1],
                    flag_invert,
                )
            else:
                # Just assign
                for idet in range(n_recv_det):
                    obs.detdata[field][idet + recv_dets[0], :] = recv_2d[idet]
            # Update per-detector flags
            dflags = {
                obs.local_detectors[recv_dets[0] + x]: defaults.det_mask_invalid
                for x in range(n_recv_det)
                if det_flags[x] != 0
            }
            obs.update_local_detector_flags(dflags)
            del recv_2d
            del recv_data
            del det_flags

    # Wait for communication
    for req in send_req:
        req.wait()
    if gcomm is not None:
        gcomm.barrier()

    # Now safe to delete our dictionary of isend buffer handles, which might include
    # temporary buffers of ranges flags.
    del send_data

    # Every process checks its local data for NaN values.  If any are found, a warning
    # is printed and the detector is cut.
    dflags = dict()
    for det in obs.local_detectors:
        nnan = np.count_nonzero(np.isnan(obs.detdata[field][det]))
        if nnan > 0:
            msg = f"{obs.name}:{det} has {nnan} NaN values.  Cutting."
            log.warning(msg)
            dflags[det] = defaults.det_mask_invalid
    obs.update_local_detector_flags(dflags)


@function_timer
def compute_boresight_pointing(
    obs,
    times_key,
    az_key,
    el_key,
    roll_key,
    quat_azel_key,
    quat_radec_key,
    position_key,
    velocity_key,
    flag_key,
    flag_mask,
):
    """Compute boresight pointing in parallel.

    Use the loaded Az / El / Roll angles to compute the boresight Az / El
    quaternions and also the RA / DEC boresight quaternions.  Also compute
    the observatory position and velocity.

    Args:
        obs (Observation):  The observation to modify.
        times_key (str):  The shared times field.
        az_key (str):  The shared Azimuth field.
        el_key (str):  The shared Elevation field.
        roll_key (str):  The shared Roll field.
        quat_azel_key (str):  The shared boresight Az/El quaternion field.
        quat_radec_key (str):  The shared boresight Ra/Dec quaternion field.
        position_key (str):  The shared telescope position field.
        velocity_key (str):  The shared telescope velocity field.
        flag_key (str):  The shared flag field.
        flag_mask (int):  The shared flag mask.

    Returns:
        None

    """
    log = Logger.get()
    timer = Timer()
    timer.start()

    comm = obs.comm
    gcomm = comm.comm_group
    rank = comm.group_rank

    site = obs.telescope.site

    # Position and velocity of the observatory are simply computed.  Only the
    # first row of the process grid needs to do this.
    position = None
    velocity = None
    if rank == 0:
        position, velocity = site.position_velocity(obs.shared[times_key])
    obs.shared[position_key].set(position, offset=(0, 0), fromrank=0)
    obs.shared[velocity_key].set(velocity, offset=(0, 0), fromrank=0)
    log.debug_rank(
        f"LoadContext {obs.name} site position/velocity took",
        comm=gcomm,
        timer=timer,
    )

    # Since this step takes some time, all processes in the group
    # contribute
    pnt_dist = distribute_uniform(obs.n_all_samples, comm.group_size)

    bore_azel = None
    bore_radec = None
    bore_flags = None

    # The sample counts and displacement for each process
    sample_count = [x.n_elem for x in pnt_dist]
    sample_displ = [x.offset for x in pnt_dist]

    # The local sample range as a slice
    slc = slice(sample_displ[rank], sample_displ[rank] + sample_count[rank], 1)

    bore_bad = np.logical_or(
        np.isnan(obs.shared[az_key].data[slc]),
        np.logical_or(
            np.isnan(obs.shared[el_key].data[slc]),
            np.isnan(obs.shared[roll_key].data[slc]),
        ),
    )
    bore_flags = np.array(obs.shared[flag_key].data[slc])
    bore_flags[bore_bad] |= flag_mask
    temp_az = np.array(obs.shared[az_key].data[slc])
    temp_el = np.array(obs.shared[el_key].data[slc])
    temp_roll = np.array(obs.shared[roll_key].data[slc])
    temp_az[bore_bad] = 0
    temp_el[bore_bad] = 0
    temp_roll[bore_bad] = 0
    bore_azel = toast.qarray.from_lonlat_angles(
        -temp_az,
        temp_el,
        temp_roll,
    )
    bore_radec = toast.coordinates.azel_to_radec(
        site,
        obs.shared[times_key].data[slc],
        bore_azel,
        use_qpoint=True,
    )

    # Gather all samples to rank zero and set the shared object elements.
    if gcomm is None:
        ftype = None
        qtype = None
    else:
        ftype = MPI.UNSIGNED_CHAR
        qtype = MPI.DOUBLE
    for name, local_data, nnz, mtype in [
        (flag_key, bore_flags, 1, ftype),
        (quat_azel_key, bore_azel, 4, qtype),
        (quat_radec_key, bore_radec, 4, qtype),
    ]:
        all_data = None
        final_data = None

        if gcomm is None:
            final_data = local_data
        else:
            counts = [x * nnz for x in sample_count]
            displ = [x * nnz for x in sample_displ]

            if rank == 0:
                # Flat buffer for the gather
                all_data = np.empty(obs.n_all_samples * nnz, dtype=local_data.dtype)
            gcomm.Gatherv(
                local_data.reshape((-1)), [all_data, counts, displ, mtype], root=0
            )

            if rank == 0:
                if nnz == 1:
                    final_data = all_data
                else:
                    final_data = all_data.reshape((-1, nnz))
        obs.shared[name].set(final_data, fromrank=0)
        del final_data
        del all_data

    del bore_flags
    del bore_azel
    del bore_radec

    log.debug_rank(
        f"LoadContext {obs.name} boresight pointing conversion took",
        comm=gcomm,
        timer=timer,
    )


def ax_name_fp_subst(var, fp_array, det=None):
    """Substitute a focalplane column value into a string.

    If det is not specified, then there should be a unique name for the whole table
    column when substituting.

    Args:
        var (str):  The input variable string
        fp_array (Table):  The focalplane astropy Table of detector properties.
        det (str):  If not None, use the value for this row of the table.

    Returns:
        (str):  The string with substitutions made.

    """
    if var is None:
        return None
    out = ""
    last = 0
    for match in re.finditer(r"(\{.*\})", var):
        out += var[last : match.start()]
        colname = match.group()
        colname = colname.replace("{", "")
        colname = colname.replace("}", "")
        if colname not in fp_array.keys():
            msg = f"Column '{colname}' not in focalplane table"
            raise RuntimeError(msg)
        if det is None:
            vals = np.unique(fp_array[colname])
            if len(vals) == 1:
                out += f"{vals[0]}"
            else:
                msg = f"Column '{colname}' has multiple values ({vals}), cannot"
                msg += " use in name"
                raise RuntimeError(msg)
        else:
            mask = fp_array[colname] == det
            out += f"{fp_array[colname][mask][0]}"
        last = match.end()
    out += var[last:]
    return out



from sotodlib.qa.quality_reports.report_data import ReportData, ReportDataConfig
from typing import Literal, Union, Dict, Any, Optional, Tuple, List
import numpy as np
import datetime as dt
import os
import json
from pathlib import Path
from sotodlib.qa.quality_reports import plots
import plotly.graph_objects as go
from jinja2 import Environment, FileSystemLoader
import sys
import yaml
import argparse
import traceback
from tqdm.auto import tqdm
from dateutil.relativedelta import relativedelta
from importlib import reload
reload(plots)

import sotodlib.site_pipeline.util as sp_util


def create_manifest(base_dir: str, output_file: str):
    manifest = []
    for parent in [os.path.join(base_dir, "weekly"), os.path.join(base_dir, "monthly")]:
        if not os.path.exists(parent):
            continue
        for folder_name in np.sort(os.listdir(parent)):
            folder_path = os.path.join(parent, folder_name)
            if not os.path.isdir(folder_path):
                continue
            index_path = os.path.join(folder_path, "report.html")
            if os.path.exists(index_path):
                label = f"{os.path.basename(parent)} / {folder_name}"
                rel_path = os.path.relpath(index_path, start=base_dir).replace(os.sep, "/")
                rel_path = f"../../{rel_path}"
                manifest.append({
                    "label": label,
                    "rel_path": rel_path
                })

    with open(output_file, "w") as f:
        json.dump(manifest, f, indent=2)


class GenerateReportConfig:
    def __init__(
        self,
        platform: Literal["satp1", "satp2", "satp3", "lat"],
        site_url: str,
        report_interval: Literal["weekly", "monthly"],
        start_time: Union[dt.datetime, float, str],
        output_root: str,
        data_config: Dict[str, Any],
        stop_time: Union[dt.datetime, float, str, None] = None,
        overwrite_html: bool = False,
        overwrite_data: bool = False,
        skip_html: bool = False,
        template_dir: Optional[str] = None,
    ) -> None:
        self.platform: Literal["satp1", "satp2", "satp3", "lat"] = platform
        self.site_url: str = site_url
        self.report_interval = report_interval

        def convert_to_datetime(
            time: Union[dt.datetime, float, str, None],
        ) -> dt.datetime:
            if isinstance(time, type(None)):
                return dt.datetime.now(tz=dt.timezone.utc)
            if isinstance(time, str):
                return dt.datetime.fromisoformat(time)
            elif isinstance(time, (int, float)):
                return dt.datetime.fromtimestamp(time)
            elif isinstance(time, dt.datetime):
                return time
            else:
                raise Exception(f"Could not convert type {type(time)} to datetime")

        self.start_time = convert_to_datetime(start_time)
        self.stop_time = convert_to_datetime(stop_time)
        self.output_root = output_root
        self.overwrite_html = overwrite_html
        self.overwrite_data = overwrite_data
        self.template_dir = template_dir
        self.data_config = data_config
        self.skip_html = skip_html

        self.time_intervals: List[Tuple[dt.datetime, dt.datetime]] = []
        if self.report_interval == "weekly":
            delta = dt.timedelta(weeks=1)
        elif self.report_interval == "monthly":
            delta = relativedelta(months=1)
        start: dt.datetime = self.start_time
        self.time_intervals = []
        now = dt.datetime.now(tz=dt.timezone.utc)
        while start < self.stop_time:
            stop: dt.datetime = start + delta
            if stop > now - dt.timedelta(hours=1):
                # Give a buffer of 1 hour to compile report for previous interval
                break
            self.time_intervals.append((start, stop))
            start += delta

    @classmethod
    def from_yaml(cls, path: str) -> "GenerateReportConfig":
        with open(path, "r") as f:
            return GenerateReportConfig(**yaml.safe_load(f))


def main(cfg: str) -> None:
    cfg = GenerateReportConfig.from_yaml(cfg)
    base_path = os.path.join(cfg.output_root, cfg.report_interval)

    for start_time, stop_time in tqdm(cfg.time_intervals):
        time_str = f"{start_time:%Y%m%d}_{stop_time:%Y%m%d}"
        subdir = os.path.join(base_path, time_str)
        report_file = os.path.join(subdir, "report.html")
        data_file = os.path.join(subdir, "data.h5")

        data_cfg = ReportDataConfig(
            start_time=start_time,
            stop_time=stop_time,
            platform=cfg.platform,
            site_url=cfg.site_url,
            **cfg.data_config,
        )

        longterm_path = None
        longterm_data = None

        if not os.path.exists(subdir):
            os.makedirs(subdir)

        try:
            if not os.path.exists(data_file) or cfg.overwrite_data:
                data: ReportData = ReportData.build(data_cfg)
                data.save(data_file)
            else:
                data = ReportData.load(data_file)

            if not os.path.exists(report_file) or cfg.overwrite_html and not cfg.skip_html:
                if data.cfg.longterm_obs_file != longterm_path:
                    longterm_data = ReportData.load(data.cfg.longterm_obs_file)
                    longterm_path = data.cfg.longterm_obs_file
                render_report(
                    cfg,
                    data,
                    report_file,
                    template_dir=cfg.template_dir,
                    longterm_data=longterm_data
                )
        except Exception as e:
            tb = ''.join(traceback.format_tb(e.__traceback__))
            print(f"Failed to generate report for {time_str}: {tb} {e}")

    create_manifest(cfg.output_root, os.path.join(cfg.output_root, "manifest.json"))


def render_report(
    cfg: GenerateReportConfig,
    data: ReportData,
    output_path: str,
    template_dir=None,
    longterm_data: Optional[ReportData] = None,
):
    if template_dir is None:
        template_dir = os.path.join(os.path.dirname(__file__), "templates")

    env = Environment(loader=FileSystemLoader(template_dir))
    template = env.get_template("report.html")

    obs_efficiency_plots = plots.wafer_obs_efficiency(data)
    source_footprint_plots = plots.source_footprints(data)
    figures: Dict[str, go.Figure] = {
        "obs_efficiency_heatmap": obs_efficiency_plots.heatmap,
        "obs_efficiency_pie": obs_efficiency_plots.pie,
        "yield_vs_pwv": plots.yield_vs_pwv(data, longterm_data=longterm_data),
        "pwv_yield_vs_time": plots.pwv_and_yield_vs_time(data),
        "source_focalplane": source_footprint_plots.focalplane,
        "source_table": source_footprint_plots.table
    }

    html_kw = dict(full_html=False, include_plotlyjs=False)
    if isinstance(data.cfg.start_time, str):
        start_time_str = data.cfg.start_time
    else:
        start_time_str = data.cfg.start_time.isoformat()
    if isinstance(data.cfg.stop_time, str):
        stop_time_str = data.cfg.stop_time
    else:
        stop_time_str = data.cfg.stop_time.isoformat()
    jinja_data = {
        "data": data,
        "report_interval": cfg.report_interval.capitalize(),
        "plots": {k: v.to_html(**html_kw) for k, v in figures.items()},
        "general_stats": {
            "Start time": dt.datetime.fromisoformat(start_time_str).strftime("%A %m/%d/%Y  %H:%M (UTC)"),
            "Stop time": dt.datetime.fromisoformat(stop_time_str).strftime("%A %m/%d/%Y  %H:%M (UTC)"),
            "Number of Observations": len([o for o in data.obs_list if o.obs_type == "obs"]),
        }
    }

    with open(output_path, "w", encoding="utf-8") as output_file:
        output_file.write(template.render(jinja_data))


def get_parser(
    parser: Optional[argparse.ArgumentParser] = None,
) -> argparse.ArgumentParser:
    if parser is None:
        p = argparse.ArgumentParser()
    else:
        p = parser
    p.add_argument(
        "cfg", type=str, help="yaml configuration file."
    )
    return p


if __name__ == '__main__':
    sp_util.main_launcher(main, get_parser)



# Copyright (c) 2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Check functionality of the muMUX crosstalk simulation

"""

import unittest

import astropy.units as u
import numpy as np

try:
    # Import sotodlib toast module first, which sets global toast defaults
    import toast
    import toast.ops
    from toast.observation import default_values as defaults

    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    toast_available = True
except ImportError as e:
    toast_available = False

from ._helpers import (calibration_schedule, close_data_and_comm,
                      simulation_test_data)

from sotodlib.toast.ops import detmap_available, pos_to_chi


class SimMuMUXCrosstalkTest(unittest.TestCase):
    def test_sim_mumux_crosstalk(self):
        if not toast_available:
            print("toast cannot be imported- skipping unit tests", flush=True)
            return

        if not detmap_available:
            print(
                "DetMap cannot be imported- skipping muMUX unit tests",
                flush=True,
            )
            return

        comm, procs, rank = toast.get_world()
        data = simulation_test_data(
            comm,
            telescope_name="LAT",
            wafer_slot="w17",
            bands="LAT_f090",
            sample_rate=10.0 * u.Hz,
            thin_fp=1,
            cal_schedule=False,
        )

        # Need to simulate atmosphere to have nonzero signal

        pointing = toast.ops.PointingDetectorSimple(
            name="det_pointing_azel",
            quats="quats_azel",
            boresight="boresight_azel",
        )

        toast.ops.SimAtmosphere(
            name="sim_atmosphere",
            add_loading=True,
            lmin_center=0.001 * u.m,
            lmin_sigma=0.0001 * u.m,
            lmax_center=1 * u.m,
            lmax_sigma=0.1 * u.m,
            xstep=5 * u.m,
            ystep=5 * u.m,
            zstep=5 * u.m,
            zmax=200 * u.m,
            gain=6e-5,
            wind_dist=3000 * u.m,
            detector_pointing=pointing,
        ).apply(data)

        # Make a copy of the data for reference

        signal0 = data.obs[0].detdata["signal"].data.copy()

        # Simple test just confirms that the operator functions

        so_ops.SimMuMUXCrosstalk(
            name="sim_mumux_crosstalk",
        ).apply(data)

        # Compare signal before and after to make sure the magnitude
        # of the effect is not crazy

        signal1 = data.obs[0].detdata["signal"].data.copy()

        rms0 = np.std(signal0)
        rmsdiff = np.std(signal0 - signal1)

        assert rms0 != 0
        assert rmsdiff / rms0 < 1e-3

        # Check that the crosstalk strength is as expected

        obs = data.obs[0]
        fp = obs.telescope.focalplane
        dets = obs.detdata["signal"].keys()
        chis = pos_to_chi(fp, dets)
        ndet = len(dets)
        nnz = len(chis)
        med = np.median(np.log10(list(chis.values())))

        print(f"ndet = {ndet}",flush=True)
        print(f"nnz = {nnz} = {nnz / ndet} ndet",flush=True)
        print(f"median(log10(chi)) = {med}",flush=True)
        assert nnz > ndet and nnz < 2 * ndet
        assert med > -3 and med < -2

        close_data_and_comm(data)

if __name__ == '__main__':
    unittest.main()



import inspect
import so3g
import numpy as np
import json
import h5py

## "temporary" fix to deal with scipy>1.8 changing the sparse setup
try:
    from scipy.sparse import csr_array
except ImportError:
    from scipy.sparse import csr_matrix as csr_array
import astropy.units as u

import flacarray

from .axisman import *
from .flagman import FlagManager

# Backwards compatibility for before skip_shape_check was added
_rm_fast_kwargs = {}
if "skip_shape_check" in inspect.signature(so3g.proj.RangesMatrix).parameters:
    _rm_fast_kwargs = {'skip_shape_check': True}

# Flatten / expand RangesMatrix

def flatten_RangesMatrix(rm):
    """Given a RangesMatrix, flatten into a small number of arrays that
    can be written to HDF or npy and then reconstructed later.

    Returns:
      Dict with arrays called 'shape', 'intervals', and 'ends'.
    """
    shape = rm.shape
    if len(shape) == 1:
        intervals = rm.ranges().reshape(-1)
        ends = np.array([len(intervals)])
        return {
            'shape': np.array(shape),
            'intervals': intervals,
            'ends': ends,
        }
    if len(shape) == 2:
        intervals = [r.ranges().reshape(-1) for r in rm.ranges]
        ends = np.cumsum([len(i) for i in intervals])
        intervals = np.hstack(intervals)
        return {
            'shape': np.array(shape),
            'intervals': intervals,
            'ends': ends,
        }
    last_end = 0
    ends, intervals = [], []
    for r in rm.ranges:
        subrm = flatten_RangesMatrix(r)
        ends.append(subrm['ends'] + last_end)
        intervals.append(subrm['intervals'])
        last_end += subrm['ends'][-1]
    return {
        'shape': np.array(shape),
        'intervals': np.hstack(intervals),
        'ends': np.hstack(ends),
    }

def expand_RangesMatrix(flat_rm):
    """Reconstruct a RangesMatrix given a dict like the one returned by
    flatten_RangesMatrix."""
    shape, intervals, ends = [
        flat_rm[k] for k in ['shape', 'intervals', 'ends']]
    if len(shape) == 1:
        r = intervals.reshape((-1, 2))
        return so3g.proj.Ranges.from_array(r, shape[0])
    ranges = []
    if shape[0] == 0:
        return so3g.proj.RangesMatrix([], child_shape=shape[1:], **_rm_fast_kwargs)
    # Otherwise non-trivial
    count = np.prod(shape[:-1])
    start, stride = 0, count // shape[0]
    for i in range(0, len(ends), stride):
        _e = ends[i:i+stride] - start
        _i = intervals[start:ends[i+stride-1]]
        ranges.append(expand_RangesMatrix(
            {'shape': shape[1:], 'intervals': _i, 'ends': _e}))
        start = ends[i+stride-1]
    return so3g.proj.RangesMatrix(ranges, child_shape=shape[1:], **_rm_fast_kwargs)

## Flatten and Expand sparse arrays
def flatten_csr_array(arr):
    """Extract information from scipy.sparse.csr_array for saving in
    hdf5 files"""
    return {
        'data': arr.data,
        'indices': arr.indices,
        'indptr': arr.indptr,
        'shape': arr.shape,
    }


def expand_csr_array(flat_arr):
    """Reconstruct csr_array from flattened versions
    """
    return csr_array( (flat_arr['data'], flat_arr['indices'],
                       flat_arr['indptr']),
                     shape=flat_arr['shape'])

# Helper functions for numpy arrays containing unicode strings; must
# be written to HDF5 as "S" type arrays.

def _retype_for_write(data):
    """Convert arrays of unicode strings to "S" arrays; pass non-string
    arrays back without doing anything..

    """
    if data.dtype.kind == 'U':
        return data.astype('S')
    return data

def _retype_for_read(data):
    """Convert arrays of type "S" to arrays of unicode strings; pass
    non-string arrays back without doing anything.

    """
    if data.dtype.kind == 'S':
        return data.astype('U')
    return data


# save/load AxisManager to HDF5.

def _safe_scalars(x):
    # Kill any np.integer, np.floating, or np.str_ vals... they don't serialize.
    if isinstance(x, list):
        return [_safe_scalars(_x) for _x in x]
    if isinstance(x, tuple):
        return tuple([_safe_scalars(_x) for _x in x])
    if isinstance(x, dict):
        return {k: _safe_scalars(v) for k, v in x.items()}
    if isinstance(x, (np.integer, np.floating, np.str_, np.bool_)):
        return x.item()
    # Must be fine then!
    return x


def _save_axisman(axisman, dest, group=None, overwrite=False, compression=None,
                  encodings=None):
    """
    See AxisManager.save.
    """
    if encodings is None:
        encodings = {}
    for k, e in encodings.items():
        if k not in axisman._assignments:
            raise ValueError(f"Encoding {e} specified for non-existent field {k}.")

    # Scheme it out...
    schema = []
    for k, assign in axisman._assignments.items():
        item = {'name': k,
                'axes': assign,
                'encoding': 'unknown',
                }
        v = axisman[k]
        e = encodings.get(k, {})   # Mark as None once consumed.

        if v is None or np.isscalar(v):
            item['encoding'] = 'scalar'
        elif isinstance(v, u.quantity.Quantity):
            if v.shape == ():
                item['encoding'] = 'scalar_quantity'
            else:
                item['encoding'] = 'quantity'
        elif isinstance(v, np.ndarray):
            item['encoding'] = 'ndarray'
            if e.get('type') is None:
                e = None
            elif e.get('type') == 'flacarray':
                item['encoding'] = 'flacarray'
                e = None
        elif isinstance(v, AxisInterface):
            item['encoding'] = 'axis'
        elif isinstance(v, AxisManager):
            item['encoding'] = 'axisman'
            if v.__class__ is AxisManager:
                item['subclass'] = 'AxisManager'
            elif v.__class__ is FlagManager:
                item['subclass'] = 'FlagManager'
                item['special_axes'] = v._dets_name, v._samps_name
            else:
                raise ValueError(f"No encoder system for {k}={v.__class__}")
            e = None
        elif isinstance(v, (so3g.RangesInt32, so3g.proj.RangesMatrix)):
            item['encoding'] = 'rangesmatrix'
        elif isinstance(v, csr_array):
            item['encoding'] = 'csrarray'
        else:
            print(v.__class__)

        if e:  # not None, not empty {}.
            raise ValueError(f"Unhandled encoding {e} for field {k}")
        schema.append(item)

    for k, v in axisman._axes.items():
        if isinstance(v, LabelAxis):
            schema.append({'name': k,
                           'encoding': 'axis',
                           'type': 'label',
                           'args': (v.name, list(v.vals))})
        elif isinstance(v, OffsetAxis):
            schema.append({'name': k,
                           'encoding': 'axis',
                           'type': 'offset',
                           'args': (v.name, v.count, v.offset, v.origin_tag)})
        elif isinstance(v, IndexAxis):
            schema.append({'name': k,
                           'encoding': 'axis',
                           'type': 'index',
                           'args': (v.name, v.count)})
        else:
            raise ValueError(f"No encoder for axis class: {v.__class__}")

    # Sanitize ...
    schema = _safe_scalars(schema)

    # Resolve the destination group.
    file_to_close = None
    if isinstance(dest, str):
        file_to_close = h5py.File(dest, 'a')
        dest = file_to_close['/']
    assert isinstance(dest, h5py.Group)  # filename or Group expected
    if group is not None:
        if group in dest:
            dest = dest[group]
        else:
            dest = dest.create_group(group)

    # Needs emptying?  This might be slower than just del dest[group],
    # but it also works for '/' or Groups passed in by reference only.
    for target in [dest, dest.attrs]:
        for k in list(target.keys()):
            if overwrite:
                del target[k]
            else:
                raise RuntimeError(
                    f'Destination group "{dest.name}" is not empty; '
                    f'pass overwite=True to clobber.')

    dest.attrs['_axisman'] = json.dumps({
        'version': 0,
        'schema': schema,
        })
    scalars = {}
    units = {}

    for item in schema:
        data = axisman[item['name']]
        if item['encoding'] == 'scalar':
            scalars[item['name']] = data
        elif item['encoding'] == 'ndarray':
            dest.create_dataset(item['name'], data=_retype_for_write(data), compression=compression)
        elif item['encoding'] == 'flacarray':
            g = dest.create_group(item['name'])
            fa = flacarray.hdf5.write_array(data, g, **encodings[item['name']].get('args', {}))
        elif item['encoding'] == 'quantity':
            dest.create_dataset(item['name'], data=_retype_for_write(data), compression=compression)
            units[item['name']] = data.unit.to_string()
        elif item['encoding'] == 'scalar_quantity':
            scalars[item['name']] = data.value
            units[item['name']] = data.unit.to_string()
        elif item['encoding'] == 'rangesmatrix':
            g = dest.create_group(item['name'])
            for k, v in flatten_RangesMatrix(data).items():
                g.create_dataset(k, data=v, compression=compression)
        elif item['encoding'] == 'csrarray':
            g = dest.create_group(item['name'])
            for k, v in flatten_csr_array(data).items():
                g.create_dataset(k, data=v, compression=compression)
        elif item['encoding'] == 'axisman':
            g = dest.create_group(item['name'])
            _save_axisman(data, g, compression=compression, encodings=encodings.get(item['name']))
        elif item['encoding'] == 'axis':
            pass #
        else:
            print(f'Unhandled {item["name"]}->{item["encoding"]}')

    if len(scalars):
        dest.attrs['_scalars'] = json.dumps(scalars)

    if len(units):
        dest.attrs['_units'] = json.dumps(units)

    if file_to_close:
        file_to_close.close()

def _get_subfields(fields, prefix):
    if fields is None:
        return None
    subfields = []
    for f in fields:
        if f.startswith(prefix + '.'):
            subfields.append(f[len(prefix)+1:])
    if len(subfields) == 0:
        return None
    return subfields

def _load_axisman(src, group=None, cls=None, fields=None):
    """
    See AxisManager.load.
    """
    if cls is None:
        cls = AxisManager

    if isinstance(src, str):
        f = h5py.File(src, 'r')
        if group is None:
            src = f
        else:
            src = f[group]
    else:
        f = None

    info = json.loads(src.attrs['_axisman'])
    assert(info['version'] == 0)
    schema = info['schema']

    scalars = {}
    if '_scalars' in src.attrs:
        scalars = json.loads(src.attrs['_scalars'])

    units = {}
    if '_units' in src.attrs:
        units = json.loads(src.attrs['_units'])

    # Reconstruct axes.
    axes = []
    for item in schema:
        if item['encoding'] == 'axis':
            if item['type'] == 'label':
                axes.append((LabelAxis(*item['args'])))
            elif item['type'] == 'offset':
                axes.append((OffsetAxis(*item['args'])))
            elif item['type'] == 'index':
                axes.append((IndexAxis(*item['args'])))
    axisman = cls(*axes)
    for item in schema:
        subfields = _get_subfields(fields, item['name'])
        if (fields is not None) and (item['name'] not in fields) and subfields is None:
            continue
        assign = [(i, a) for i, a in enumerate(item.get('axes', [])) if a is not None]
        if item['encoding'] == 'axis':
            pass
        elif item['encoding'] == 'scalar':
            axisman.wrap(item['name'], scalars[item['name']])
        elif item['encoding'] == 'ndarray':
            axisman.wrap(item['name'], _retype_for_read(src[item['name']][:]), assign)
        elif item['encoding'] == 'flacarray':
            d = flacarray.hdf5.read_array(src[item['name']])
            axisman.wrap(item['name'], d, assign)
        elif item['encoding'] == 'quantity':
            axisman.wrap(item['name'], _retype_for_read(src[item['name']][:]) << u.Unit(units[item['name']]), assign)
        elif item['encoding'] == 'scalar_quantity':
            axisman.wrap(item['name'], scalars[item['name']] << u.Unit(units[item['name']]), assign)
        elif item['encoding'] == 'axisman':
            x = _load_axisman(src[item['name']], fields=subfields)
            if item['subclass'] == 'FlagManager':
                x = FlagManager.promote(x, *item['special_axes'])
            axisman.wrap(item['name'], x, restrict_in_place=True)
        elif item['encoding'] == 'rangesmatrix':
            x = src[item['name']]
            rm_flat = {k: x[k][:] for k in ['shape', 'intervals', 'ends']}
            axisman.wrap(item['name'], expand_RangesMatrix(rm_flat), assign)
        elif item['encoding'] == 'csrarray':
            x = src[item['name']]
            csr_flat = {k: x[k][:] for k in ['shape', 'data', 'indices', 'indptr']}
            axisman.wrap(item['name'], expand_csr_array(csr_flat), assign)
        else:
            print('No decoder for:', item)

    return axisman



""" Utility functions for querying and/or working with the G3tHK databases
"""

import numpy as np
import logging

from sotodlib.io.g3thk_db import G3tHk, HKFiles, HKAgents, HKFields

logger = logging.getLogger(__name__)


def pysmurf_monitor_control_list(agent, start=None, stop=None, HK=None):
    """Return list of stream_ids controlled by a pysmurf-monitor agent

    Arguments
    ----------
    agent: HKAgent instance or instance_id
    start: timestamp
        if agent is instance_id, used to find relevant agent instance
    stop: timestamp
        if agent is instance_id, used to find relevant agent instance
    HK: G3tHK Archive
        if agent is instance_id, used to find relevant agent instance

    Returns
    -------
    stream_ids: list of stream_ids monitored by agent
    """
    if isinstance(agent, str):
        if start is None or stop is None:
            raise ValueError(
                "start and stop are required when agent is the" " instance id"
            )
        if HK is None:
            raise ValueError("need database to search if agent is instance id")
        agent_list = HK.get_db_agents(agent, start, stop)
        if len(agent_list) == 0: 
            logger.warning(f"Agent {agent} not running between {start} and {stop}")
            return np.array([], dtype='<U8') 
        return np.unique(
                np.concatenate([pysmurf_monitor_control_list(agent) for agent in agent_list])
        )
    stream_ids = []
    for field in agent.fields:
        if not "_meta" in field.field:
            continue
        splits = field.field.split(".")
        sid = [x for x in splits if "_meta" in x][0].strip("_meta")
        stream_ids.append(sid)
    return np.unique(stream_ids)


def check_was_streaming(stream_id, start, stop, cfgs=None, HK=None, servers=None):
    """Query the HK database to see if a specific stream_id was
    streaming during a time range.

    Arguments
    ----------
    stream_id: string
    start: timestamp
    stop: timestamp
    cfgs: optional, dictionary
        configuration dictionary with finalization informtion. Format matches
        what is described in G3tSmurf.from_configs. If not given, HK and servers
        arguments are required.
    servers: optional, list
        list of dictionaries containing the server to instance id mapping as
        described in G3tSmurf.from_configs. In not given the cfgs dictionary is
        used instead.
    HK: optional, G3tHK instance
        if not given, loaded using configs dictionary

    Returns
    -------
    check: boolean
        if true, pysmurf-monitor has recorded the g3 stream was open
        during the time between start and stop.

    Raises
    ------
    ValueError if stream_id is found in multiple pysmurf-monitors
    """

    if HK is None:
        HK = G3tHk.from_configs(cfgs)
    if servers is None:
        servers = cfgs["finalization"]["servers"]
    assert HK.get_last_update() >= stop

    agents = None

    for server in servers:
        pm = server.get("pysmurf-monitor")
        if pm is None:
            continue
        alist = HK.get_db_agents(pm, start, stop)
        sids = np.unique([pysmurf_monitor_control_list(agent) for agent in alist])
        if np.any([stream_id == s for s in sids]):
            if agents is not None:
                logger.error(f"found {stream_id} in multiple pysmurf-montiors")
                raise ValueError(
                    f"found {stream_id} in multiple "
                    f"pysmurf-montiors, do not know which one to "
                    "use"
                )
            agents = alist

    if agents is None:
        ## found no agents, either not streaming or db not updated
        logger.warning(
            f"Found no pysmurf monitor agents monitoring {stream_id}"
            " during this time"
        )
        return False

    fields = []
    search = f"{stream_id}_meta.AMCcSmurfProcessorSOStreamopen_g3stream"

    for agent in agents:
        idx = np.where([search in field.field for field in agent.fields])[0]
        if len(idx) == 0:
            continue
        fields.append(agent.fields[idx[0]])

    if len(fields) == 0:
        logger.warning(
            f"Found no fields with key {search}. Was metadata" "reading out correctly?"
        )
        return False

    data = HK.load_data(fields)
    if len(data) != 1:
        logger.error("Data returned from field list has more than one key")

    k, val = data.popitem()
    msk = np.all([val[0] >= start, val[0] <= stop], axis=0)

    return np.any(val[1][msk])



from typing import Optional, Dict
from dataclasses import dataclass, fields

import so3g
from so3g.proj import Ranges
from spt3g import core
import itertools
import numpy as np
from scipy.interpolate import interp1d
from scipy.signal import convolve
from tqdm.auto import tqdm
import os
import logging
import sys
import shutil
import yaml
import datetime as dt
from zipfile import ZipFile, ZIP_DEFLATED
import sotodlib
from sotodlib.site_pipeline.util import init_logger
from .datapkg_utils import walk_files


log = logging.getLogger('bookbinder')
if not log.hasHandlers():
    init_logger('bookbinder')

class TimingSystemOff(Exception):
    """Exception raised when we try to bind books where the timing system is found to be off and the books have imprecise timing counters"""
    pass

class NoScanFrames(Exception):
    """Exception raised when we try and bind a book but the SMuRF file contains not Scan frames (so no detector data)"""
    pass

MAX_DROPPED_SAMPLES = 200
class BadTimeSamples(Exception):
    """Exception raised when there are drops in the time samples in the 
    UFM timestreams"""
    pass

class NoHKFiles(Exception):
    """Exception raised when we cannot find any HK data around the book time"""
    pass

class NoMountData(Exception):
    """Exception raised when we cannot find mount data"""
    pass

class NoHWPData(Exception):
    """Exception raised when we cannot find HWP data"""
    pass

class DuplicateAncillaryData(Exception):
    """Exception raised when we find the HK data has copies of the same timestamps"""
    pass

class NonMonotonicAncillaryTimes(Exception):
    """Exception raised when we find the HK data has timestamps that are not strictly increasing monotonically"""
    pass

class BookDirHasFiles(Exception):
    """Exception raised when files already exist in a book directory"""
    pass

def setup_logger(logfile=None):
    """
    This setups up a logger for bookbinder. If a logfile is passed, it will
    write to that file as well as stdout. It is useful to create a one-off
    logger here instead of using `getLogger` because it allows us to set
    a separate log-file for each bookbinder instance.
    """
    fmt = '%(asctime)s: %(message)s (%(levelname)s)'
    log = logging.Logger('bookbinder', level=logging.DEBUG)

    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(fmt)
    ch.setLevel(logging.INFO)
    ch.setFormatter(formatter)
    log.addHandler(ch)

    if logfile is not None:
        ch = logging.FileHandler(logfile)
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(formatter)
        log.addHandler(ch)

    return log

def get_frame_iter(files):
    """
    Returns a continuous iterator over frames for a list of files.
    """
    return itertools.chain(*[core.G3File(f) for f in files])

def close_writer(writer):
    """
    Closes out a G3FileWriter with an end-processing frame. If None is passed,
    this will not do anything.
    """
    if writer is None:
        return
    writer(core.G3Frame(core.G3FrameType.EndProcessing))

def next_scan(it):
    """
    Returns the next Scan frame, along with any intermediate frames for an
    iterator.
    """
    interm_frames = []
    for frame in it:
        if frame.type == core.G3FrameType.Scan:
            return frame, interm_frames
        interm_frames.append(frame)
    return None, interm_frames

class HkDataField:
    """
    Class containing HK Data for a single field.

    Args
    -----
    instance_id: str
        Instance id of the agent producing the data.
    feed: str
        Feed name for hk data feed.
    field: str
        Field name for hk data feed.

    Attributes
    -----------
    times: np.ndarray
        HK sample timestamps.
    data: np.ndarray
        HK sample data.
    finalized: bool
        True if datas has been processed and finalized.
    """
    def __init__(self, instance_id: str, feed: str, field: str):
        self.instance_id = instance_id
        self.feed = feed
        self.field = field

        self.times =[]
        self.data = []
        self.finalized = False

    def __len__(self):
        return len(self.times)

    @property
    def addr(self):
        """Returns full address of field"""
        return f"{self.instance_id}.{self.feed}.{self.field}"

    def process_frame(self, frame):
        """Update data based on G3Frame"""
        address = frame['address']  # "<site>.<instance_id>.feeds.<feed_name>""
        spl = address.split('.')
        instance_id, feed = spl[1], spl[3]
        if instance_id != self.instance_id or feed != self.feed:
            return

        for block in frame['blocks']:
            if self.field not in block:
                continue
            self.times.append(np.array(block.times) / core.G3Units.s)
            self.data.append(block[self.field])

    def finalize(self, drop_duplicates=False, require_monotonic_times=True):
        """Finalize data, and store in numpy array"""
        self.times = np.hstack(self.times, dtype=np.float64)
        self.data = np.hstack(self.data)
        self.finalized = True

        # Check for duplicates
        clean_times, idxs = np.unique(self.times, return_index=True)
        if len(self.times) != len(clean_times):
            if not drop_duplicates:
                raise DuplicateAncillaryData(
                    f"HK data from {self.addr} has" 
                    " duplicate timestamps"
                )
            else:
                log.warning(
                    f"HK data from {self.addr} has duplicate timestamps"
                )
            self.times = self.times[idxs]
            self.data = self.data[idxs]
        if not np.all(np.diff(self.times)>0):
            bad = np.sum( np.diff(self.times) <= 0)
            msg = f"Times from {self.addr} have {bad} samples that are " \
                "not increasing"
            if require_monotonic_times:
                raise NonMonotonicAncillaryTimes(msg)
            else:
                log.warning(msg)
        
@dataclass
class HkData:
    """
    Class containing HkData for bookbinding, including ACU and HWP data.
    """
    az: Optional[HkDataField] = None
    el: Optional[HkDataField] = None
    boresight: Optional[HkDataField] = None
    corotator_enc: Optional[HkDataField] = None
    az_mode: Optional[HkDataField] = None
    hwp_freq: Optional[HkDataField] = None

    @classmethod
    def from_dict(cls, d: Dict[str, str]):
        """
        Creates HkData object from a dict of field addresses. Addresses must be formatted like ``<instance_id>.<feed>.<field>``.
        Keys of dict must be valide fields of HkData class, such as ``az`` or ``el``.
        """
        kw = {}
        for k, v in d.items():
            try:
                instance, feed, field = v.split('.')
            except Exception as exc:
                raise ValueError(f"Could not parse field: {v}. "
                                 "Must be formatted <instance_id>.<feed>.<field>") from exc
            kw[k] = HkDataField(instance, feed, field)
        return cls(**kw)

    def process_frame(self, frame):
        """Processes G3frame, and updates relevant HkDataFields"""
        for fld in fields(self):
            f = getattr(self, fld.name)
            if isinstance(f, HkDataField):
                f.process_frame(frame)

    def finalize(self, drop_duplicates=True, require_monotonic_times=True):
        """Finalizes HkDatafields"""
        for fld in fields(self):
            f = getattr(self, fld.name)
            if isinstance(f, HkDataField):
                f.finalize(
                    drop_duplicates=drop_duplicates,
                    require_monotonic_times=require_monotonic_times,
                )

class AncilProcessor:
    """
    Processor for ancillary (ACU) data

    Params
    --------
    files : list
        List of HK files to process.
    book_id: str
        ID of book being bound.
    hk_fields: dict
        Dictionary of fields corresponding to relevant HK Data. See the HkData
        class for what housekeeping fields are allowed.  For example::

            >> hk_fields = {
                'az': 'acu.acu_udp_stream.Corrected_Azimuth',
                'el': 'acu.acu_udp_stream.Corrected_Elevation',
                'boresight': 'acu.acu_udp_stream.Corrected_Boresight',
                # corotator_enc: acu.acu_status.Corotator_current_position # (For LAT)
                'az_mode':  'acu.acu_status.Azimuth_mode',
                'hwp_freq': 'hwp-bbb-e1.HWPEncoder.approx_hwp_freq',
            }

    Attributes
    -----------
    hkdata : HkData
        Class containing relevant Hk data for the duration of the book.
    times : np.ndarray
        Timestamps for anc data. This will be populated after preprocess.
    anc_frame_data : List[G3TimestreamMap]
        List of G3TimestreamMaps saved for each bound frame. This will be
        populated on bind and should be used to add copies of the anc data to
        the detector frames.
    """
    def __init__(self, files, book_id, hk_fields: Dict, 
                 drop_duplicates=False, require_hwp=True, 
                 require_acu=True, require_monotonic_times=True, 
                 log=None
                 ):
        self.hkdata: HkData = HkData.from_dict(hk_fields)

        self.files = files
        self.anc_frame_data = None
        self.out_files = []
        self.book_id = book_id
        self.preprocessed = False
        self.drop_duplicates = drop_duplicates
        self.require_hwp = require_hwp
        self.require_acu = require_acu
        self.require_monotonic_times = require_monotonic_times
        if log is None:
            self.log = logging.getLogger('bookbinder')
        else:
            self.log = log

        if len(self.files) == 0:
            if self.require_acu or self.require_hwp:
                raise NoHKFiles("No HK files specified for book")
            self.log.warning("No HK files found for book")
            for fld in ['az', 'el', 'boresight', 'corotator_enc','az_mode', 'hwp_freq']:
                setattr(self.hkdata, fld, None)

        if self.require_acu and self.hkdata.az is None:
            self.log.warning("No ACU data specified in hk_fields!")

        if self.require_hwp and self.hkdata.hwp_freq is None:
            self.log.warning("No HWP Freq data is specified in hk_fields.")

    def preprocess(self):
        """
        Preprocesses HK data and populates the `data` and `times` objects.
        """
        if self.preprocessed:
            return

        self.log.info("Preprocessing HK Data")
        frame_iter = get_frame_iter(self.files)

        for fr in frame_iter:
            if fr['hkagg_type'] != 2:
                continue
            self.hkdata.process_frame(fr)

        # look for ACU fields that are configured but not found in HK data files
        # will not check fields that are not in the configuration file
        for fld in ['az', 'el', 'boresight', 'corotator_enc']:
            f = getattr(self.hkdata, fld)
            if f is not None:
                if self.require_acu and len(f) == 0:
                    raise NoMountData(
                        f"Did not find ACU data in {self.files} for {fld}",
                    )
                elif len(f) == 0:
                    ## requiring Az data is fails and we didn't find any
                    self.log.warning(
                        f"Did not find ACU data for {fld}. Bypassed because "
                        "require_acu is false"
                    )
                    setattr(self.hkdata, fld, None)

        # look for HWP data if HWP fields are in configuration file
        if self.hkdata.hwp_freq is not None:
            if self.require_hwp and len(self.hkdata.hwp_freq) == 0:
                raise NoHWPData(
                    f"Did not find HWP data in {self.files}",
                )
            elif len(self.hkdata.hwp_freq) == 0:
                ## requiring HWP data is false and we didn't find any
                self.log.warning(
                    f"Did not find HWP data in data. Bypassed because "
                    "require_hwp is false"
                )
                self.hkdata.hwp_freq = None
        
        self.hkdata.finalize(
            drop_duplicates=self.drop_duplicates,
            require_monotonic_times=self.require_monotonic_times,
        )
        self.preprocessed = True

    def bind(self, outdir, times, frame_idxs, file_idxs):
        """
        Binds ancillary data.

        Params
        ----------
        outdir : path
            Path where files should be written
        times : np.ndarray
            Array of timestamps to write to book
        frame_idxs : np.ndarray
            Array mapping sample (in times) to output frame index
        file_idxs : np.ndarray
            Array mapping output frame idx to output file idx
        """
        self.log.info("Binding ancillary data")

        # Handle file writers
        writer = None
        cur_file_idx = None
        out_files = []

        def validate_mount_field(hk_field: HkDataField, max_dt=None):
            m = (times[0] <= hk_field.times) & (hk_field.times <= times[-1])
            if m.sum() < 2:
                raise NoMountData(
                    f"No mount data overlapping with detector data: {hk_field.addr}"
                )
            if max_dt is not None:
                _max_dt = np.max(np.diff(hk_field.times[m]))
                if _max_dt > max_dt:
                    raise NoMountData(
                        f"Max data spacing {_max_dt}s is higher than {max_dt}s for {hk_field.addr}. "
                        "Interpolation may be questionable."
                    )

        # go through and interpolate ACU times to detector times
        acu_interp_data = {}
        for fld in ['az', 'el', 'boresight', 'corotator_enc']:
            f = getattr(self.hkdata, fld)
            if f is not None:
                try:
                    validate_mount_field(f, max_dt=10)
                    acu_interp_data[fld] = np.interp(
                        times, f.times, f.data
                    )
                except NoMountData as e:
                    if self.require_acu:
                        raise e
                    else:
                        self.log.warning(e)
                        acu_interp_data[fld] = None
            else: 
                acu_interp_data[fld] = None

        az = acu_interp_data['az']
        el = acu_interp_data['el']
        boresight = acu_interp_data['boresight']
        corotator_enc = acu_interp_data['corotator_enc']

        anc_frame_data = []
        for oframe_idx in np.unique(frame_idxs):
            # Update file writer if starting a new output file
            if file_idxs[oframe_idx] != cur_file_idx:
                close_writer(writer)
                cur_file_idx = file_idxs[oframe_idx]
                fname = os.path.join(outdir, f'A_ancil_{cur_file_idx:0>3}.g3')
                out_files.append(fname)
                writer = core.G3Writer(fname)

            m = frame_idxs == oframe_idx
            ts = times[m]

            oframe = core.G3Frame(core.G3FrameType.Scan)

            i0, i1 = np.where(m)[0][[0, -1]]
            oframe['sample_range'] = core.G3VectorInt([int(i0), int(i1+1)])
            oframe['book_id'] = self.book_id

            anc_data = core.G3TimesampleMap()
            anc_data.times = core.G3VectorTime(ts * core.G3Units.s)
            if az is not None:
                anc_data['az_enc'] = core.G3VectorDouble(az[m])
                anc_data['el_enc'] = core.G3VectorDouble(el[m])
            if boresight is not None:
                anc_data['boresight_enc'] = core.G3VectorDouble(boresight[m])
            if corotator_enc is not None:
                anc_data['corotator_enc'] = core.G3VectorDouble(corotator_enc[m])
            oframe['ancil'] = anc_data
            writer(oframe)
            anc_frame_data.append(anc_data)

            self.add_acu_summary_info(oframe, ts[0], ts[-1])

        # Save this to be added to detector files
        self.anc_frame_data = anc_frame_data
        self.out_files = out_files

    def add_acu_summary_info(self, frame, t0, t1):
        """
        Adds ACU summary information to a G3Frame. This will add the following
        info if data is present in the HK dataset::
            - azimuth_mode: (str)
                Azimuth_mode, pulled from the ACU summary data. `ProgramTrack`
                means that this frame contains scan data, and `Preset` means the
                telescope is slewing.
            - azimuth_velocity_mean / azimuth_velocity_std: (float / float)
                Mean and standard deviation of the az velocity (deg / sec)
            - elevation_velocity_mean / elevaction_velocity_std: (float / float)
                Mean and standard deviation of the el velocity (deg / sec)

        Params
        ----------
        frame : G3Frame
            Frame to add data to
        t0 : float
            Start time of frame (unix time), inclusive
        t1 : float
            Stop time of frame (unix time), inclusive
        """
        az_mode = self.hkdata.az_mode
        az = self.hkdata.az
        el = self.hkdata.el
        if az_mode is not None:
            m = (t0 <= az_mode.times) & (az_mode.times <= t1)
            if not np.any(m):
                frame['azimuth_mode'] = 'None'
            elif 'ProgramTrack' in az_mode.data[m]:
                frame['azimuth_mode'] = 'ProgramTrack' # Scanning
            else:
                frame['azimuth_mode'] = 'Preset'  # Slewing

        if az is not None:
            m = (t0 <= az.times) & (az.times <= t1)
            if m.sum() >= 2:
                dt = np.diff(az.times[m]).mean()
                az_vel = np.diff(az.data[m]) / dt
                frame['azimuth_velocity_mean'] = np.mean(az_vel)
                frame['azimuth_velocity_stdev'] = np.std(az_vel)
        for k in ['azimuth_velocity_mean', 'azimuth_velocity_stdev']:
            if k not in frame:
                frame[k] = np.nan

        if el is not None:
            m = (t0 <= el.times) & (el.times <= t1)
            if m.sum() >= 2:
                dt = np.diff(el.times[m]).mean()
                el_vel = np.diff(el.data[m]) / dt
                frame['elevation_velocity_mean'] = np.mean(el_vel)
                frame['elevation_velocity_stdev'] = np.std(el_vel)
        for k in ['elevation_velocity_mean', 'elevation_velocity_stdev']:
            if k not in frame:
                frame[k] = np.nan

class SmurfStreamProcessor:
    def __init__(self, obs_id, files, book_id, readout_ids,
                 log=None, allow_bad_timing=False, min_ctime=None,
                 max_ctime=None):
        self.files = files
        self.obs_id = obs_id
        self.stream_id = None
        self.times = None
        self.frame_idxs = None
        self.nchans = None
        self.nframes = None
        self.bias_names = None
        self.primary_names = None
        self.timing_paradigm = None
        self.session_id = None
        self.slow_primary = None
        self.sostream_version = None
        self.readout_ids = readout_ids
        self.out_files = []
        self.book_id = book_id
        self.allow_bad_timing = allow_bad_timing
        self.min_ctime = min_ctime
        self.max_ctime = max_ctime

        if log is None:
            self.log = logging.getLogger('bookbinder')
        else:
            self.log = log

    def preprocess(self):
        """
        Extracts file times, nchans, and nframes from file list
        """
        if self.times is not None:  # Already preprocessed
            return

        self.log.info(f"Preprocessing smurf obsid {self.obs_id}")

        self.nframes = 0
        ts = []
        smurf_frame_counters = [] # smurf frame-counter
        fc_idx = None
        frame_idxs = []
        frame_idx = 0
        timing = True
        for frame in get_frame_iter(self.files):
            if frame.type != core.G3FrameType.Scan:
                continue

            # Populate attributes from the first scan frame
            if self.nchans is None:
                self.nchans = len(self.readout_ids)
                self.primary_names = frame['primary'].names
                fc_idx = list(self.primary_names).index("FrameCounter")
                self.bias_names = frame['tes_biases'].names
                self.timing_paradigm = frame['timing_paradigm']
                self.session_id = frame['session_id']
                if 'slow_primary' in frame:
                    self.slow_primary = frame['slow_primary']
                self.sostream_version = frame['sostream_version']
                self.stream_id = frame['sostream_id']

            good, t = get_frame_times(frame, self.allow_bad_timing)
            timing = timing and good
            ts.append(t)
            smurf_frame_counters.append(frame['primary'].data[fc_idx])
            frame_idxs.append(np.full(len(t), frame_idx, dtype=np.int32))

            self.nframes += 1
            frame_idx += 1

        if len(ts) == 0:
            raise NoScanFrames(f"{self.obs_id} has no detector data")

        self.times = np.hstack(ts)
        self.smurf_frame_counters = np.hstack(smurf_frame_counters)
        self.frame_idxs = np.hstack(frame_idxs)
 
        if self.min_ctime is None:
            self.min_ctime = np.min(self.times)
        if self.max_ctime is None:
            self.max_ctime = np.max(self.times)
        
        msk = np.all([
            self.times >= self.min_ctime,
            self.times <= self.max_ctime,
        ], axis=0)
        if np.any(np.diff(self.times[msk]) <= 0):
            raise BadTimeSamples(
                f"{self.obs_id} has time samples not increasing"
            )
        
        timing = timing and (not self.timing_paradigm=='Low Precision')
        
        if (not self.allow_bad_timing) and (not timing):
            raise TimingSystemOff(
                f"Observation {self.obs_id} does not have high precision timing"
                " information. Pass `allow_bad_timing=True` to bind anyway"
            )
        
        # If low-precision, we need to linearize timestamps in order for
        # bookbinder to work properly
        if not timing:
            self.log.warning(
                "Timestamps are Low Precision, linearizing from frame-counter"
            )
            dt, offset = np.polyfit(self.smurf_frame_counters, self.times, 1)
            self.times = offset + dt * self.smurf_frame_counters

    def bind(self, outdir, times, frame_idxs, file_idxs, pbar=False, ancil=None,
             atol=1e-4):
        """
        Binds SMuRF data
        
        Params
        ------ 
        outdir : str
            Output directory to put bound files
        times : np.ndarray
            Full array of timestamps that should be contained in the books
        frame_idxs : np.ndarray
            Output frame idx for each specified timestamp
        file_idxs : np.ndarray
            Output file idx for each specified output frame
        pbar : bool or tqdm.tqdm
            If True, will create a new progress bar. If False, will disable.
            If a progress bar is passed, will use that.
        ancil : AncilProcessor
            Ancillary processor object (must be already bound). Ancil data will
            be copied into the output frames.
        atol : float
            Absolute tolerance between smurf-timestamps and book-timestamps.
            Samples mapped to times that are further away than atol will
            be considered unmapped.
        """
        if pbar is True:
            pbar = tqdm(total=self.nframes)
        elif pbar is False:
            pbar = tqdm(total=self.nframes, disable=True)

        pbar.set_description(f"Binding {self.stream_id}")

        self.log.info(f"Binding smurf obsid {self.obs_id}")

        # Here `times` is the full array of times in the book (gapless reference
        # times) and `self.times`` is timestamps from the pre-processed L2 data
        # (may contain gaps)
        sample_map = find_ref_idxs(times, self.times)
        mapped = np.abs(times[sample_map] - self.times) < atol

        oframe_idxs = frame_idxs[sample_map]  # out-frame idx for each in sample
        oframe_idxs[~mapped] = -1 
        _, offsets = np.unique(frame_idxs, return_index=True)
        # Sample idx within each out-frame for every input sample
        out_offset_idxs = sample_map - offsets[frame_idxs[sample_map]]

        iframe_idxs = self.frame_idxs  #in-frame idx for each in sample
        _, offsets = np.unique(self.frame_idxs, return_index=True)
        # Sample idx within each in-frame for every input sample
        in_offset_idxs = np.arange(len(self.times)) - offsets[self.frame_idxs]

        # Mask to use to separate out non-resonator signal from resonator signal
        tracked_chans = np.array(['NONE' not in r for r in self.readout_ids], dtype=bool)
        readout_id_arr = np.array(self.readout_ids)

        # Handle file writers
        writer = None
        cur_file_idx = None
        out_files = []

        inframe_iter = get_frame_iter(self.files)
        iframe, interm_frames = next_scan(inframe_iter)
        iframe_idx = 0
        oframe_num = 0
        pbar.update()

        for oframe_idx in np.unique(frame_idxs):
            # Update writer
            if file_idxs[oframe_idx] != cur_file_idx:
                close_writer(writer)
                cur_file_idx = file_idxs[oframe_idx]
                fname = os.path.join(
                    outdir, f'D_{self.stream_id}_{cur_file_idx:0>3}.g3')
                out_files.append(fname)
                writer = core.G3Writer(fname)

            # Initialize stuff
            m = frame_idxs == oframe_idx
            nsamp = np.sum(m)
            ts = times[m]
            data = np.zeros((self.nchans, nsamp), dtype=np.int32)
            biases = np.zeros((len(self.bias_names), nsamp), dtype=np.int32)
            primary = np.zeros((len(self.primary_names), nsamp), dtype=np.int64)
            filled = np.zeros(nsamp, dtype=bool)
            
            # Loop through in_frames filling current out_frame
            while True:
                # First, write any intermediate frames like observation and wiring
                for fr in interm_frames:
                    if 'frame_num' in fr:
                        del fr['frame_num']
                    fr['frame_num'] = oframe_num  # Update this so they remain ordered
                    oframe_num += 1
                    writer(fr)

                m = (oframe_idxs == oframe_idx) & (iframe_idxs == iframe_idx)
                outsamps = out_offset_idxs[m]
                insamps = in_offset_idxs[m]

                # Below is equivalent to:
                #    >> data[:, outsamps] = iframe['data'][:, insamps]
                #    >> ...
                # However it is much faster to copy arrays by mapping contiguous
                # chunks to contiguous chunks using slices, like:
                #    >> arr_out[:, o0:o1] = arr_in[:, i0:i1]
                # since numpy does not need to create a temporary copy of the
                # data, and can just do a direct mem-map. This speeds up binding
                # by a factor of ~4.
                #
                # Here we are splitting outsamps and insamps into a list
                # of ranges where both arrays are contiguous. Then we loop
                # through each sub-range and copy data via slicing.
                split_idxs = 1 + np.where(
                    (np.diff(outsamps) > 1) | (np.diff(insamps) > 1))[0]
                outsplits = np.split(outsamps, split_idxs)
                insplits = np.split(insamps, split_idxs)
                for i in range(len(outsplits)):
                    if len(insplits[i]) == 0:
                        continue

                    in0, in1 = insplits[i][0], insplits[i][-1] + 1
                    out0, out1 = outsplits[i][0], outsplits[i][-1] + 1

                    data[:, out0:out1] = iframe['data'].data[:, in0:in1]
                    biases[:, out0:out1] = iframe['tes_biases'].data[:, in0:in1]
                    primary[:, out0:out1] = iframe['primary'].data[:, in0:in1]
                    filled[out0:out1] = 1

                # If there are any remaining samples in the next in_frame, pull it and repeat
                if np.any((oframe_idxs == oframe_idx) & (iframe_idxs > iframe_idx)):
                    iframe, interm_frames = next_scan(inframe_iter)
                    iframe_idx += 1
                    pbar.update()
                    continue
                else:
                    break

            # Interpolate data where there are gaps
            if np.all(~filled):
                self.log.error(
                    f"No samples properly mapped in oframe frame {oframe_idx}!"
                    "Cannot properly interpolate for this frame."
                )
                raise ValueError(f"Cannot finish binding {self.obs_id}")
            elif np.any(~filled):
                self.log.warning(
                    f"{np.sum(~filled)} missing samples in out-frame {oframe_idx}"
                )
                # Missing samples at the beginning / end of a frame will be
                # filled with the first / last sample in the frame
                fill_value = (data[:, filled][:, 0], data[:, filled][:, -1])
                data[:, ~filled] = interp1d(
                    ts[filled], data[:, filled], axis=1, assume_sorted=True,
                    kind='linear', fill_value=fill_value, bounds_error=False
                )(ts[~filled])

            m = frame_idxs == oframe_idx
            i0, i1 = np.where(m)[0][[0, -1]]

            oframe = core.G3Frame(core.G3FrameType.Scan)

            if ancil is not None:
                t0, t1 = ts[0], ts[-1]
                oframe['ancil'] = ancil.anc_frame_data[oframe_idx]
                ancil.add_acu_summary_info(oframe, t0, t1)

            oframe['book_id'] = self.book_id 
            oframe['sample_range'] = core.G3VectorInt([int(i0), int(i1+1)])
            oframe['flag_smurfgaps'] = core.G3VectorBool(~filled)

            ts = core.G3VectorTime(ts * core.G3Units.s)
            oframe['signal'] = so3g.G3SuperTimestream(readout_id_arr[tracked_chans], ts, data[tracked_chans, :])
            oframe['untracked'] = so3g.G3SuperTimestream(readout_id_arr[~tracked_chans], ts, data[~tracked_chans, :])
            oframe['primary'] = so3g.G3SuperTimestream(self.primary_names, ts, primary)
            oframe['tes_biases'] = so3g.G3SuperTimestream(self.bias_names, ts, biases)
            oframe['stream_id'] = self.stream_id
            oframe['frame_num'] = oframe_num

            oframe_num += 1
            writer(oframe)
        
        close_writer(writer)
        self.out_files = out_files

        # In case there were remaining frames left over
        pbar.update(self.nframes - iframe_idx - 1)
        if pbar.n >= pbar.total:
            pbar.close()

class BookBinder:
    """
    Class for combining smurf and hk L2 data to create books containing detector
    timestreams.

    Parameters
    ----------
    book : sotodlib.io.imprinter.Books
        Book object to bind
    obsdb : dict
        Result of imprinter.get_g3tsmurf_obs_for_book. This should be a dict
        from obs-id to G3tSmurf Observations.
    filedb : dict
        Result of imprinter.get_files_for_book. This should be a dict from
        obs-id to the list of smurf-files for that observation.
    hkfiles : list
        List of HK files to process.
    max_samps_per_frame : int
        Max number of samples per frame. This will be used to split frames
        when the ACU data isn't present or cannot be used.
    max_file_size : int
        Max file size in bytes. This will be used to rotate files.
    readout_ids : dict, optional
        Dict of readout_ids to use for each stream_id. If provided, these
        will be used to set the `names` in the signal frames. If not provided,
        names will be taken from the input frames.
    ignore_tags : bool, optional
        if true, will ignore tags if the level 2 observations have unmatched
        tags
    ancil_drop_duplicates: bool, optional
        if true, will drop duplicate timestamp data from ancillary files. added
        to deal with an occassional hk aggregator error where it is picking up
        multiple copies of the same data
    require_acu: bool, optional
        if true, will throw error if we do not find Mount data
    require_monotonic_times: bool, optional
        if true, will throw error if we ever see timestamps not increasing or going backwards
    require_hwp: bool, optional
        if true, will throw error if we do not find HWP data
    allow_bad_time: bool, optional
        if not true, books will not be bound if the timing systems signals are not found. 
    min_ctime: float, optional
        if not None, will cut book to this minimum ctime
    max_ctime: float optional
        if not None, will cut book to this maximum ctime
    
    Attributes
    -----------
    ancil : AncilProcessor
        Processor for ancillary data
    streams : dict
        Dict of SmurfStreamProcessor objects, keyed by stream_id
    times : np.ndarray
        Array of times for all samples in the book
    frame_idxs : np.ndarray
        Array of output frame indices for all samples in the book
    file_idxs : np.ndarray
        Array of output file indices for all output frames in the book
    """
    def __init__(self, book, obsdb, filedb, data_root, readout_ids, 
                outdir, hk_fields, max_samps_per_frame=50_000, max_file_size=1e9, ignore_tags=False, ancil_drop_duplicates=False, 
                require_hwp=True, require_acu=True,
                require_monotonic_times=True,
                allow_bad_timing=False,
                min_ctime=None, max_ctime=None):
        self.filedb = filedb
        self.book = book
        self.data_root = data_root
        self.hk_root = os.path.join(data_root, 'hk')
        self.meta_root = os.path.join(data_root, 'smurf')
        
        self.obsdb = obsdb
        self.outdir = outdir

        assert book.schema==0, "obs/oper books only have schema=0"

        self.max_samps_per_frame = max_samps_per_frame
        self.max_file_size = max_file_size
        self.ignore_tags = ignore_tags
        self.allow_bad_timing = allow_bad_timing
        self.min_ctime = min_ctime
        self.max_ctime = max_ctime

        if os.path.exists(outdir):
            # don't count hidden files, possibly from NFS processes
            nfiles = len([f for f in os.listdir(outdir) if f[0] != '.'])
            if nfiles > 1:
                raise BookDirHasFiles(
                    f"Output directory {outdir} contains files. Delete to retry"
                      " bookbinding"
                )

            elif nfiles == 1:
                assert os.listdir(outdir)[0] == 'Z_bookbinder_log.txt', \
                    f"only acceptable file in new book path {outdir} is " \
                    " Z_bookbinder_log.txt"

        else:
            os.makedirs(outdir)

        logfile = os.path.join(outdir, 'Z_bookbinder_log.txt')
        self.log = setup_logger(logfile)

        try:
            self.hkfiles = get_hk_files(
                self.hk_root, 
                book.start.timestamp(),
                book.stop.timestamp()
            )
        except NoHKFiles as e:
            if require_hwp or require_acu:
                self.log.error(
                    "HK files are required if we require ACU or HWP data"
                )
                raise e
            self.log.warning(
                "Found no HK files during book time, binding anyway because "
                "require_acu and require_hwp are False"
            )
            self.hkfiles = []            

        self.ancil = AncilProcessor(
            self.hkfiles, 
            book.bid, 
            hk_fields,
            log=self.log, 
            drop_duplicates=ancil_drop_duplicates,
            require_hwp=require_hwp,
            require_acu=require_acu,
            require_monotonic_times=require_monotonic_times,
        )
        self.streams = {}
        for obs_id, files in filedb.items():
            obs = self.obsdb[obs_id]
            stream_id = obs.stream_id
            if not obs.timing and not self.allow_bad_timing:
                raise TimingSystemOff(
                    f"Observation {obs_id} does not have high precision timing "
                    "information. Pass `allow_bad_timing=True` to bind anyway"
                )
            self.streams[stream_id] = SmurfStreamProcessor(
                obs_id, files, book.bid, readout_ids[obs_id], log=self.log,
                allow_bad_timing=self.allow_bad_timing, 
                min_ctime=self.min_ctime, max_ctime=self.max_ctime,
            )

        self.times = None
        self.frame_idxs = None
        self.file_idxs = None
        self.meta_files = None

    def set_min_max_ctime(self):
        """Function to be run after stream.preprocess is finished to set the ctimes.
        Splitting this out because it is useful for debugging purposes as well
        """   
        for stream in self.streams.values():
            assert stream.times is not None, "All streams must be preprocessed"
        t0 = np.max([s.times[0] for s in self.streams.values()])
        if self.min_ctime is not None:
            assert self.min_ctime >= t0, \
                f"{self.min_ctime} is less than the first time found in"\
                f" the detector data {t0}"
            self.log.warning(
                f"Over-riding minimum ctime from {t0} to {self.min_ctime}"
            )
        else:
            self.min_ctime = t0

        t1 = np.min([s.times[-1] for s in self.streams.values()])
        if self.max_ctime is not None:
            assert self.max_ctime <= t1, \
                f"{self.max_ctime} is greater than the last time found in"\
                f" the detector data {t1}"
            self.log.warning(
                f"Over-riding maximum ctime from {t1} to {self.max_ctime}"
            )
        else:
            self.max_ctime = t1

    def preprocess(self):
        """
        Runs preprocessing steps for the book. Preprocesses smurf and ancillary
        data. Creates full list of book-times, the output frame idx for each
        sample, and the output file idx for each frame.
        """
        if self.times is not None:
            return

        for stream in self.streams.values():
            stream.preprocess()
        self.set_min_max_ctime()

        # prioritizes the last stream
        # implicitly assumes co-sampled (this is where we could throw errors
        # after looking for co-sampled data)
        ts, _ = fill_time_gaps(stream.times) 
        m = (self.min_ctime <= ts) & (ts <= self.max_ctime)
        ts = ts[m]

        self.ancil.preprocess()

        # Divide up frames, only look within detector data and +/-30 seconds
        frame_splits = find_frame_splits(self.ancil, ts[0]-30, ts[-1]+30)

        if frame_splits is None:
            frame_idxs = np.arange(len(ts)) // self.max_samps_per_frame
        else:
            frame_idxs = np.digitize(ts, frame_splits)
            frame_idxs -= frame_idxs[0]
            new_frame_idxs = frame_idxs.copy()
            # Divide up frames that are too long
            for fidx in np.unique(frame_idxs):
                m = frame_idxs == fidx
                new_frame_idxs += np.cumsum(m) // self.max_samps_per_frame
            frame_idxs = new_frame_idxs

        # Divide up files
        samp_size = 4 # bytes
        max_chans = np.max([s.nchans for s in self.streams.values()])
        totsize = samp_size * max_chans * np.arange(len(ts))
        file_idxs = []
        for fr in np.unique(frame_idxs):
            idx = np.where(frame_idxs == fr)[0][-1]
            file_idxs.append(totsize[idx] // self.max_file_size)
        file_idxs = np.array(file_idxs, dtype=int)

        self.times = ts
        self.frame_idxs = frame_idxs
        self.file_idxs = file_idxs

        self.check_timesamples()
        self.log.info("Finished preprocessing data")

    def check_timesamples(self, atol=1e-4):
        """
        Checks for missing timesamples in individual streams relative to the 
        book times. Makes sure individual readout slots haven't dropped too many
        points
        """
        if self.times is None:
            raise ValueError(
                "Preprocess must have been run to check_timesamples"
            )
        
        self.dropped = {}
        for u, s in self.streams.items():
            sample_map = find_ref_idxs(self.times, s.times)
            mapped = np.abs(self.times[sample_map] - s.times) < atol
            diffs = np.diff(sample_map[mapped])
            idx = np.where( diffs>1)[0]
            self.dropped[u] = sum( [diffs[i]-1 for i in idx] )
        
        if np.all( [x==0 for x in self.dropped.values()] ):
            ## no dropped samples from any slot
            return

        ## do not change the format of these messages without also changing the 
        ## autofixing behavior that is built off these messages.
        msg = '\n'.join([
            f"\t{self.streams[u].obs_id}: {x}" for u, x in self.dropped.items()
        ])
        if np.any( [x>MAX_DROPPED_SAMPLES for x in self.dropped.values()]):
            if (not self.allow_bad_timing):
                raise BadTimeSamples(
                    f"Streams have more than {MAX_DROPPED_SAMPLES} time samples"
                    f" missing. Pass `allow_bad_timing=True` to bind anyway. "
                    "Missing samples:\n" + msg
                )
            else:
                self.log.warning(
                    f"Streams have more than {MAX_DROPPED_SAMPLES} time samples"
                    f" missing. Missing Samples: \n" + msg
                )
        else:
            self.log.warning(
                f"Streams have time samples missing. Missing Samples: \n" + msg
            )


    def copy_smurf_files_to_book(self):
        """
        Copies smurf ancillary files to an operation book.
        """
        if self.book.type != 'oper':
            return

        self.log.info("Copying smurf ancillary files to book")

        files = []
        for obs in self.obsdb.values():
            files.extend(get_smurf_files(obs, self.meta_root))
        
        smurf_dirname = 'Z_smurf'
        os.makedirs(os.path.join(self.outdir, smurf_dirname), exist_ok=True)

        meta_files = {}
        for f in files:
            relpath = os.path.join(smurf_dirname, os.path.basename(f))
            dest = os.path.join(self.outdir, relpath)
            self.log.info(f"Copying to {dest}")
            shutil.copyfile(f, dest)

            if f.endswith('iv_analysis.npy'):
                meta_files['iv'] = relpath
            elif f.endswith('bg_map.npy'):
                meta_files['bgmap'] = relpath
            elif f.endswith('bias_step_analysis.npy'):
                meta_files['bias_steps'] = relpath
            elif f.endswith('take_noise.npy'):
                meta_files['noise'] = relpath

        self.meta_files = meta_files

    def write_M_files(self, telescope, tube_config):
        # write M_book file
        m_book_file = os.path.join(self.outdir, "M_book.yaml")
        book_meta = {}
        book_meta["book"] = {
            "type": self.book.type,
            "schema_version": self.book.schema,
            "book_id": self.book.bid,
            "finalized_at": dt.datetime.utcnow().isoformat(),
        }
        book_meta["bookbinder"] = {
            "codebase": sotodlib.__file__,
            "version": sotodlib.__version__,
            # leaving this in but KH doesn't know what it's supposed to be for
            "context": "unknown", 
        }
        with open(m_book_file, "w") as f:
            yaml.dump(book_meta, f)
        
        mfile = os.path.join(self.outdir, "M_index.yaml")
        with open(mfile, "w") as f:
            yaml.dump(
                self.get_metadata(
                    telescope=telescope,
                    tube_config=tube_config,
                ), f
            )

    def get_metadata(self, telescope=None, tube_config={}):
        """
        Returns metadata dict for the book
        """
        self.preprocess()

        meta = {}
        meta['book_id'] = self.book.bid
        meta['type'] = self.book.type

        meta['start_time'] = float(self.times[0])
        meta['stop_time'] = float(self.times[-1])
        meta['n_frames'] = len(np.unique(self.frame_idxs))
        meta['n_samples'] = len(self.times)
        meta['session_id'] = self.book.bid.split('_')[1]
        meta['filled_samples'] = {k:int(x) for k,x in self.dropped.items()}

        sample_ranges = []
        for file_idx in np.unique(self.file_idxs):
            fr_idxs = np.where(self.file_idxs == file_idx)[0]
            i0 = int(np.where(self.frame_idxs == fr_idxs[0])[0][0])
            i1 = int(np.where(self.frame_idxs == fr_idxs[-1])[0][-1])
            sample_ranges.append([i0, i1+1])
        meta['sample_ranges'] = sample_ranges

        if telescope is None:
            self.log.warning(
                "telescope not explicitly defined. guessing from book"
            )
            meta['telescope'] = self.book.tel_tube[:3].lower()
        else: 
            meta['telescope'] = telescope

        if 'tube_slot' not in tube_config:
            self.log.warning("tube_slot key missing from tube_config. guessing")
        meta['tube_slot'] = tube_config.get(
            'tube_slot',
            self.book.tel_tube.lower().replace("sat","satst")[3:]
        )
        meta['tube_flavor'] = tube_config.get('tube_flavor')
        meta['wafer_slots'] = tube_config.get('wafer_slots')

        detsets = []
        tags = []

        # build detset list in same order as slots
        meta['stream_ids'] = self.book.slots.split(',')
        for sid in meta['stream_ids']:
            detsets.append(
                [obs.tunesets[0].name for _,obs in self.obsdb.items() 
                    if obs.stream_id == sid ][0]
            )
        # just append all tags, order doesn't matter
        for _, g3tobs in self.obsdb.items():
            tags.append(g3tobs.tag)
        meta['detsets'] = detsets

        hwp_freq = self.ancil.hkdata.hwp_freq
        meta['hwp_freq_mean'] = None
        meta['hwp_freq_stdev'] = None
        t0, t1 = self.times[0], self.times[-1]
        if hwp_freq is not None:
            m = (t0 < hwp_freq.times) & (hwp_freq.times < t1)
            if m.any():
                meta['hwp_freq_mean'] = float(np.mean(hwp_freq.data[m]))
                meta['hwp_freq_stdev'] = float(np.std(hwp_freq.data[m]))
        
        az = self.ancil.hkdata.az
        meta['az_speed_mean'] = None
        meta['az_speed_stdev'] = None
        if az is not None:
            m = (t0 < az.times) & (az.times <= t1)
            if np.sum(m) >= 2:
                dt = np.diff(az.times[m]).mean()
                az_speed = np.abs(np.diff(az.data[m]) / dt)
                meta['az_speed_mean'] = float(np.mean(az_speed))
                meta['az_speed_stdev'] = float(np.std(az_speed))

        # make sure all tags are the same for obs in the same book
        tags = list(set(tags))
        if not self.ignore_tags:
            assert len(tags) == 1
        else:
            tags = [','.join(tags)]
        tags = tags[0].split(',')
        # book should have at least one tag
        assert len(tags) > 0
        meta['subtype'] = tags[1] if len(tags) > 1 else ""
        # sanitize rest of tags
        meta['tags'] = [t.strip() for t in tags[2:] if t.strip() != '']
        
        if (self.book.type == 'oper') and self.meta_files:
            meta['meta_files'] = self.meta_files
        return meta

    def bind(self, pbar=False):
        """
        Binds data.

        Params
        ---------
        pbar : bool
            If True, will enable a progress bar.
        """
        self.preprocess()
        
        self.log.info(f"Binding data to {self.outdir}")
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)

        # Copy smurf ancillary files if they exist
        self.copy_smurf_files_to_book()

        # Bind Ancil Data
        self.ancil.bind(self.outdir, self.times, self.frame_idxs, self.file_idxs)
        
        tot = np.sum([s.nframes for s in self.streams.values()])
        pbar = tqdm(total=tot, disable=(not pbar))
        for stream in self.streams.values():
            stream.bind(self.outdir, self.times, self.frame_idxs,
                        self.file_idxs, pbar=pbar, ancil=self.ancil)

        self.log.info("Finished binding data. Exiting.")
        return True

class TimeCodeBinder:
    """Class for building the timecode based books, smurf, stray, and hk books. 
    These books are built primarily just by copying specified files from level 
    2 locations to new locations at level 2.
    """

    def __init__(
        self, book, timecode, indir, outdir, file_list=None, 
        ignore_pattern=None,
    ):
        self.book = book
        self.timecode = timecode
        self.indir = indir
        self.outdir = outdir
        self.file_list = file_list
        if ignore_pattern is not None:
            self.ignore_pattern = ignore_pattern
        else:
            self.ignore_pattern = []
        
        if book.type == 'smurf' and book.schema > 0:
            self.compress_output = True
        else:
            self.compress_output = False    

    def get_metadata(self, telescope=None, tube_config={}):
        return {
            "book_id": self.book.bid,
            # dummy start and stop times
            "start_time": float(self.timecode) * 1e5,
            "stop_time": (float(self.timecode) + 1) * 1e5,
            "telescope": telescope,
            "type": self.book.type,
        }
    
    def write_M_files(self, telescope, tube_config):
        # write M_book file
        
        book_meta = {}
        book_meta["book"] = {
            "type": self.book.type,
            "schema_version": self.book.schema,
            "book_id": self.book.bid,
            "finalized_at": dt.datetime.utcnow().isoformat(),
        }
        book_meta["bookbinder"] = {
            "codebase": sotodlib.__file__,
            "version": sotodlib.__version__,
            # leaving this in but KH doesn't know what it's supposed to be for
            "context": "unknown", 
        }
        if self.compress_output:
            with ZipFile(self.outdir, mode='a') as zf:
                zf.writestr("M_book.yaml", yaml.dump(book_meta))
        else:
            m_book_file = os.path.join(self.outdir, "M_book.yaml")
            with open(m_book_file, "w") as f:
                yaml.dump(book_meta, f)
        
        index = self.get_metadata(
            telescope=telescope,
            tube_config=tube_config,
        )
        if self.compress_output:
            with ZipFile(self.outdir, mode='a') as zf:
                zf.writestr("M_index.yaml", yaml.dump(index))
        else:
            mfile = os.path.join(self.outdir, "M_index.yaml")
            with open(mfile, "w") as f:
                yaml.dump(index, f)

    def bind(self, pbar=False):
        if self.compress_output:
            if self.file_list is None:
                self.file_list = walk_files(self.indir, include_suprsync=True)
                ignore = shutil.ignore_patterns(*self.ignore_pattern)
                to_ignore = ignore("", self.file_list)
                self.file_list = sorted(
                    [f for f in self.file_list if f not in to_ignore]
                )
            with ZipFile(self.outdir, mode='x') as zf:
                for f in self.file_list:
                    relpath = os.path.relpath(f, self.indir)
                    zf.write(f, arcname=relpath, compress_type=ZIP_DEFLATED)
        elif self.file_list is None:
            shutil.copytree(
                self.indir,
                self.outdir,
                ignore=shutil.ignore_patterns(
                    *self.ignore_pattern,
                ),
            )
        else:
            if not os.path.exists(self.outdir):
                os.makedirs(self.outdir)
            for f in self.file_list:
                relpath = os.path.relpath(f, self.indir)
                path = os.path.join(self.outdir, relpath)
                base, _ = os.path.split(path)
                if not os.path.exists(base):
                    os.makedirs(base)
                shutil.copy(f, os.path.join(self.outdir, relpath))

def fill_time_gaps(ts):
    """
    Fills gaps in an array of timestamps.

    Parameters
    -------------
    ts : np.ndarray
        List of timestamps of length `n`, potentially with gaps
    
    Returns
    --------
    new_ts : np.ndarray
        New list of timestamps of length >= n, with gaps filled.
    mask : np.ndarray
        Returns a mask that tells you which elements of `new_ts` are
        taken from real data and which are interpolated. `~mask` gives
        the indices of the interpolated elements.
    """
    # Find indices where gaps occur and how long each gap is
    dts = np.diff(ts)
    dt = np.median(dts)
    missing = np.round(dts/dt - 1).astype(int)
    total_missing = int(np.sum(missing))

    # Create new array with the correct number of samples
    new_ts = np.full(len(ts) + total_missing, np.nan)

    # Insert old timestamps into new array with offsets that account for gaps
    offsets = np.concatenate([[0], np.cumsum(missing)])
    i0s = np.arange(len(ts))
    new_ts[i0s + offsets] = ts

    # Use existing data to interpolate and fill holes
    m = np.isnan(new_ts)
    xs = np.arange(len(new_ts))
    interp = interp1d(xs[~m], new_ts[~m])
    new_ts[m] = interp(xs[m])

    return new_ts, ~m

_primary_idx_map = {}
def get_frame_times(frame, allow_bad_timing=False):
    """
    Returns timestamps for a G3Frame of detector data.

    Parameters
    --------------
    frame : G3Frame
        Scan frame containing detector data
    allow_bad_timing: bool, optional
        if not true, raises an error if it finds data with imprecise timing

    Returns
    --------------
    high_precision : bool
        If true, timestamps are computed from timing counters. If not, they are
        software timestamps
    
    timestamps : np.ndarray
        Array of timestamps (sec) for samples in the frame

    """
    if len(_primary_idx_map) == 0:
        for i, name in enumerate(frame['primary'].names):
            _primary_idx_map[name] = i
        
    c0 = frame['primary'].data[_primary_idx_map['Counter0']]
    c2 = frame['primary'].data[_primary_idx_map['Counter2']]

    counters = np.all( np.diff(c0)!=0 ) and np.all( np.diff( c2 )!=0)

    if counters:
        return True, counters_to_timestamps(c0, c2)
    elif allow_bad_timing:
        return False, np.array(frame['data'].times) / core.G3Units.s
    else:
        ## don't change this error message. used in Imprinter CLI
        raise TimingSystemOff("Timing counters not incrementing")

def split_ts_bits(c):
    """
    Split up 64 bit to 2x32 bit
    """
    NUM_BITS_PER_INT = 32
    MAXINT = (1 << NUM_BITS_PER_INT) - 1
    a = (c >> NUM_BITS_PER_INT) & MAXINT
    b = c & MAXINT
    return a, b

def counters_to_timestamps(c0, c2):
    s, ns = split_ts_bits(c2)

    # Add 20 years in seconds (accounting for leap years) to handle
    # offset between EPICS time referenced to 1990 relative to UNIX time.
    c2 = s + ns*1e-9 + 5*(4*365 + 1)*24*60*60
    ts = np.round(c2 - (c0 / 480000) ) + c0 / 480000
    return ts

def find_ref_idxs(refs, vs):
    """
    Creates a mapping from a list of timestamps (vs) to a list of reference
    timestamps (refs). Returns an index-map of shape `vs.shape`, that maps each
    timestamp of the array `vs` to the closest timestamp in the array `refs`.

    This assumes that `refs` and `vs` are sorted in ascending order.

    Parameters
    ----------
    refs : array_like
        List of reference timestamps
    vs : array_like
        List of timestamps

    Returns
    -------
    idxs : array_like
        Map of shape `vs.shape` that maps each element of `vs` to the closest
        element in `refs`.
    """
    # Find the indices of the samples in the list of timestamps (vs)
    # that are closest to the reference timestamps
    idx = np.searchsorted(refs, vs, side='left')
    idx = np.clip(idx, 1, len(refs)-1)
    # shift indices to the closest sample
    left = refs[idx-1]
    right = refs[idx]
    idx -= vs - left < right - vs
    return idx

def get_hk_files(hkdir, start, stop, tbuff=10*60):
    """
    Gets HK files for dat between start and stop dirHWP files that may have data
    between start and stop time
    """
    files = []
    for subdir in os.listdir(hkdir):
        try:
            tcode = int(subdir)
        except:
            continue

        if not start//1e5 - 1 <= tcode <= stop//1e5 + 1:
            continue

        subpath = os.path.join(hkdir, subdir)
        files.extend([os.path.join(subpath, f) for f in os.listdir(subpath)])

    files = np.array(sorted(files))
    file_times = np.array(
        [int(os.path.basename(f).split('.')[0]) for f in files]
    )

    m = (start-tbuff <= file_times) & (file_times < stop+tbuff)
    if not np.any(m):
        check = np.where( file_times <= start )
        if len(check) < 1 or len(check[0]) < 1:
            raise NoHKFiles(
                f"Cannot find HK files between {start} and {stop}"
            )
        fidxs = [check[0][-1]]
        m[fidxs] = 1
    else:
        fidxs = np.where(m)[0]
    # Add files before and after for good measure
    
    i0, i1 = fidxs[0], fidxs[-1]
    if i0 > 0:
        m[i0 - 1] = 1
    if i1 < len(m) - 1:
        m[i1 + 1] = 1

    return files[m].tolist()

def locate_scan_events(
        times, az, 
        vel_thresh=0.01, # mount noise for satps is 0.015 deg/s level
        min_gap=200, 
        filter_window=100
    ):
    """
    Locate places where the azimuth velocity changes sign, including starts and
    stops. These locations are where we should start determining the scan 
    framing.

    Parameters
    ----------
    times : ndarray float
        times 
    az: ndarray float
        azimuth positions
    vel_thresh : float, optional
        threshold for what is considered stopped
    min_gap : int, optional
        Length of gap (in samples) longer than which events are considered separate

    Returns
    -------
    events : list
        Full list containing all zero-crossings, starts, and stops that should become frame edges
    """

    if len(az) < 1:
        return []

    offset = 0
    vel = np.diff(az)/np.diff(times)

    if filter_window is not None:
        win = np.hanning(filter_window) / np.sum(np.hanning(filter_window))
        vel = convolve(vel, win, mode='same')[filter_window:-filter_window]
        offset = filter_window

    ## find places with "zero" velocity
    zero_vel = np.abs(vel) < vel_thresh
    if np.all(zero_vel):
        return []
    
    zeros = Ranges.from_mask(zero_vel)
    zeros.close_gaps(min_gap)
    
    ## find places where velocity changes sign in case the velocity
    ## is so fast it never gets close enough to zero
    x = np.where( np.sign(vel) > 0 )[0]
    y = np.where( np.diff(x) > 1 )[0]
    cross = Ranges.zeros_like(zeros)
    for z in y:
        cross.add_interval( x[z]+1, x[z]+2)
    x = np.where( np.sign(vel) < 0 )[0]
    y = np.where( np.diff(x) > 1 )[0]
    for z in y:
        cross.add_interval( x[z]+1, x[z]+2)
    
    zeros = zeros + cross
    events = []
    
    for c in zeros.ranges():   
        # if zero period is longer than min_gap, it's a start or stop add each side to the list
        if c[1] - c[0] > min_gap:
            if c[0] != 0:
                events.append( c[0] )
            if c[1] != len(vel):
                events.append( c[1] )
        # otherwise, it's a zero crossing, add mean
        else:
            events.append( int(round( sum(c)/2 )) )
    
    return np.array(events, dtype='int')+offset

def find_frame_splits(ancil, t0=None, t1=None):
    """
    Determines timestamps of frame-splits from ACU data. If it cannot determine
    frame-splits, returns None.

    Arguments
    ----------
    ancil: AncillaryProcesser
    t0: float (optional)
        start time to analyze ACU behavior 
    t1: float (optional)
        stop time to analyze ACU behavior
    """
    az = ancil.hkdata.az
    if az is None:
        return None

    if t0 is None:
        t0 = az.times[0]
    if t1 is None:
        t1 = az.times[-1]

    msk = np.all(
        [az.times >= t0, az.times <= t1],
        axis=0
    )
    idxs = locate_scan_events(az.times[msk], az.data[msk], filter_window=100)
    return az.times[msk][idxs]

def get_smurf_files(obs, meta_path, all_files=False):
    """
    Returns a list of smurf files that should be copied into a book.

    Parameters
    ------------
    obs : G3tObservations
        Observation to pull files from
    meta_path : path
        Smurf metadata path
    all_files : bool
        If true will return all found metadata files

    Returns
    -----------
    files : List[path]
        List of copyable files
    """

    def copy_to_book(file):
        if all_files:
            return True
        return file.endswith('npy')

    tscode = int(obs.action_ctime//1e5)
    files = []

    # check adjacent folders in case action falls on a boundary
    for tc in [tscode-1, tscode, tscode + 1]:
        action_dir = os.path.join(
            meta_path,
            str(tc),
            obs.stream_id,
            f'{obs.action_ctime}_{obs.action_name}'
        )

        if not os.path.exists(action_dir):
            continue

        for root, _, fs in os.walk(action_dir):
            files.extend([os.path.join(root, f) for f in fs])

    return [f for f in files if copy_to_book(f)]



import unittest
import numpy as np
from sotodlib import core, tod_ops
from sotodlib.tod_ops import t2pleakage
from sotodlib.hwp import hwp

def quick_dets_axis(n_det):
    return core.LabelAxis('dets', ['det%04i' % i for i in range(n_det)])

def gen_1_over_f(n_samp, dt, fknee):
    tmp = np.random.random(size=n_samp) * 2 - 1
    S = np.fft.rfft(tmp)

    fknee = 1
    freqs = np.fft.rfftfreq(n_samp, dt)
    freqs[0] = freqs[1]
    fil = 1 + (freqs / fknee) ** -2
    S = S * fil
    s = np.fft.irfft(S)
    s /= np.max(np.abs(s))
    return s

def quick_tod(n_det, n_samp, dt=.005, fknee=0.5, fhwp=2.0, t2q=1e-3, t2u=4e-3):
    dets = quick_dets_axis(n_det)
    tod = core.AxisManager(
        quick_dets_axis(n_det),
        core.OffsetAxis('samps', n_samp))
    timestamps = np.arange(n_samp) * dt
    
    tod.wrap_new('timestamps', shape=('samps', ))[:] = 1800000000. + timestamps
    tod.wrap_new('signal', shape=('dets', 'samps'), dtype='float32')
    
    _t = tod.timestamps - tod.timestamps[0]
    tod.wrap_new('hwp_angle',  shape=('samps', ))[:] = (2*np.pi*fhwp*_t + 1.32 + 1e-5*np.random.randn(len(_t))) % (np.pi*2)
    
    for di, det in enumerate(tod.dets.vals):
        atm = gen_1_over_f(n_samp,dt, fknee)
        atm += np.ptp(atm) * 5
        tod.signal[di] = atm * (1 + t2q*np.cos(4*tod.hwp_angle) + t2u*np.sin(4*tod.hwp_angle))
    
    return tod

class LeakageSubtractionTest(unittest.TestCase):
    "Test the temperature-to-polarization leakage subtaction"
    def test_leakage_subtraction_joint(self):
        #prepare tod
        n_det = 10
        n_samp = 200*600
        dt=.005
        fknee=0.5
        fhwp=2.0
        t2q=1e-3
        t2u=4e-3
        
        tod = quick_tod(n_det=n_det, n_samp=n_samp, 
                        dt=dt, fknee=fknee, fhwp=fhwp,
                        t2q=t2q, t2u=t2u)
        
        hwp.demod_tod(tod)
        oman = t2pleakage.get_t2p_coeffs(tod)

        self.assertTrue(np.all(np.isclose(oman.lamQ, t2q, atol=oman.lamQ_error*5, rtol=0)))
        self.assertTrue(np.all(np.isclose(oman.lamU, t2u, atol=oman.lamU_error*5, rtol=0)))

    def test_leakage_subtraction(self):
        #prepare tod
        n_det = 10
        n_samp = 200*600
        dt=.005
        fknee=0.5
        fhwp=2.0
        t2q=1e-3
        t2u=4e-3
        
        tod = quick_tod(n_det=n_det, n_samp=n_samp, 
                        dt=dt, fknee=fknee, fhwp=fhwp,
                        t2q=t2q, t2u=t2u)
        
        hwp.demod_tod(tod)
        oman = t2pleakage.get_t2p_coeffs(tod, joint_fit=False)

        self.assertTrue(np.all(np.isclose(oman.coeffsQ, t2q, atol=oman.errorsQ*5, rtol=0)))
        self.assertTrue(np.all(np.isclose(oman.coeffsU, t2u, atol=oman.errorsU*5, rtol=0)))
        
if __name__ == '__main__':
    unittest.main()


import h5py
import numpy as np
import os
import yaml
from copy import deepcopy
from typing import Optional, Dict, List
from dataclasses import dataclass
from tqdm.auto import tqdm
import logging
import argparse
import time

from sotodlib.coords import det_match, optics
from sotodlib import core
from sotodlib.core.metadata import ManifestDb
from sotodlib.io.metadata import write_dataset


logger = logging.getLogger("update_det_match")
if len(logger.handlers) == 0:
    fmt = logging.Formatter("%(asctime)s %(levelname)s %(name)s : %(message)s")
    handler = logging.StreamHandler()
    handler.setFormatter(fmt)
    logger.addHandler(handler)
logger.setLevel(logging.INFO)


@dataclass
class UpdateDetMatchesConfig:
    """
    Configuration for update script

    Args
    ------
    results_path: str
        Path to directory where results such as matches, manifestdbs, and
        h5 files will be stored.
    context_path: str
        Path to context file. This must contain detset and det_cal metadata.
    site_pipeline_root: str
        Path to root of site-pipeline-configs. If ``$SITE_PIPELINE_CONFIG_DIR``
        is set in the environment, that will be used as the default.
    wafer_map_path: str
        Path to wafer-map to be used to find det-match solution files. If not specified,
        defaults to ``<site_pipeline_root>/shared/detmapping/wafer_map.yaml``.
    freq_offset_range_args: Optional[Tuple[float, float, float]]
        If this is not None, for each match, we will scan over a range of
        freq-offsets to determine the optimal offset to use. If set, must
        contain a tuple of floats, containing ([start,] stop, [step,]) that will
        be passed directly to ``np.arange``. If it is None, will just run with
        the match with freq_offset_mhz=0.
    match_pars: Optional[Dict]
        If not None, will be passed directly to ``det_match.MatchParams`` that
        is used by the det-match function.
    detset_meta_name: str
        Name of the metadata entry in the context that contains detset info.
    detcal_meta_name: str
        Name of the metadata entry in the context that contains det_cal info.
    show_pb: bool
        Will show progress bar when scanning freq-offset.
    apply_solution_pointing: bool
        If True, pointing information computed from design-detector positions
        will be used in the ``merged`` detset of the match.
    write_relpath: bool
        If True, will use the relative path to the h5 file (relative to the db
        path) when writing to the manifestdb
    solution_type: str
        Type of solutions to use. Must be one of ['kaiwen_handmade',
        'resonator_set'].  If 'kaiwen_handmade', will use the handmade solutions
        from Kaiwen pulled from the wafer_map file in the site-pipeline-configs.
        If `resonator_set`, must also specify the ``resonator_set_dir`` to pull
        solutions from.
    resonator_set_dir: Optional[str]
        If ``solution_type`` is 'resonator_set', this must be specified and
        contain the path to the resonator-set solutions. This directory must
        have a res-set npy file for each stream_id that is expected in the
        matching, formatted like ``<resonator_set_dir>/<stream_id>.npy``, which
        contains the result from ``np.save(fname, match.merged.as_array())``.
    time_before_cache_failure: float
        Time in seconds before a failed detset will be added to the cache. This
        is to prevent new detsets still acquiring data from being added right
        away.
    
    Attributes
    -------------
    freq_offsets : Optional[np.ndarray]
        If not None, contains freq_offsets determined by
        ``freq_offset_range_args`` which will be scanned over.
    """
    results_path: str
    context_path: str
    site_pipeline_root:str = os.environ.get('SITE_PIPELINE_CONFIG_DIR')
    wafer_map_path: Optional[str] = None
    freq_offset_range_args: Optional[tuple[float, float, float]] = (-4, 4, 0.3)
    match_pars: Optional[Dict] = None
    detset_meta_name : str = 'smurf'
    detcal_meta_name: str = 'det_cal'
    show_pb: bool = False
    apply_solution_pointing: bool = True
    write_relpath: bool = True
    solution_type: str = 'kaiwen_handmade'
    resonator_set_dir: Optional[str] = None
    time_before_cache_failure: float = float(3600 * 24 * 7)

    def __post_init__(self):
        if self.site_pipeline_root is None:
            raise ValueError("Must set site_pipeline_root, or SITE_PIPELINE_CONFIG_DIR env var")

        if self.wafer_map_path is None:
            self.wafer_map_path = os.path.join(
                self.site_pipeline_root, 'shared/detmapping/wafer_map.yaml')
        
        if self.freq_offset_range_args is not None:
            self.freq_offsets = np.arange(*self.freq_offset_range_args)
        else:
            self.freq_offsets = None

        if self.match_pars is None:
            self.match_pars = {}

        if not os.path.exists(self.results_path):
            raise FileNotFoundError(f"Results dir does not exist: {self.results_path}")
        
        allowed_solution_types = ['kaiwen_handmade', 'resonator_set']
        if self.solution_type not in allowed_solution_types:
            raise ValueError(
                f"Solution type ({self.solution_type}) must be a member of: {allowed_solution_types}")
        
        if self.solution_type == 'resonator_set':
            if self.resonator_set_dir is None:
                raise ValueError("Must specify resonator_set_dir for solution_type='resonator_set'")


class Runner:
    def __init__(self, cfg: UpdateDetMatchesConfig):
        self.cfg = cfg
        self.ctx = core.Context(cfg.context_path)
        self.detset_db = None
        self.detcal_db = None
        with open(self.cfg.wafer_map_path, 'r') as f:
            self.wafer_map = yaml.safe_load(f)
        
        self.ufm_to_fp_file = os.path.join(
            cfg.site_pipeline_root, 'shared/focalplane/ufm_to_fp.yaml')

        for d in self.ctx['metadata']:
            if 'name' in d:
                entry_name = d['name']
            elif 'label' in d:
                entry_name = d['label']
            else:
                continue
            if entry_name == cfg.detset_meta_name:
                self.detset_db = core.metadata.ManifestDb(d['db'])
            elif entry_name == cfg.detcal_meta_name:
                self.detcal_db = core.metadata.ManifestDb(d['db'])

        if self.detset_db is None:
            raise Exception(
                f"Could not find detset metadata entry with name: {cfg.detset_meta_name}")
        if self.detcal_db is None:
            raise Exception(
                f"Could not find detcal metadata entry with name: {cfg.detcal_meta_name}")
        
        self.failed_detset_cache_path = os.path.join(cfg.results_path, 'failed_detsets.yaml')
        self.match_dir = os.path.join(cfg.results_path, 'matches')
        if not os.path.exists(self.match_dir):
            os.mkdir(self.match_dir)

    def get_remaining_detsets(self) -> List[str]:
        detsets_all = set(self.detset_db.get_entries(['dataset'])['dataset'])
        failed_detsets = set(get_failed_detsets(self.failed_detset_cache_path))
        finished_detsets = set([os.path.splitext(f)[0] for f in os.listdir(self.match_dir)])
        remaining_detsets = list(detsets_all - failed_detsets - finished_detsets)
        return remaining_detsets

def load_solution_set(runner: Runner, stream_id: str, wafer_slot=None):
    cfg = runner.cfg
    if cfg.solution_type == 'kaiwen_handmade':
        sol_file = os.path.join(
            os.path.dirname(runner.cfg.wafer_map_path),
            runner.wafer_map[stream_id]['solution']
        )
        teltype = runner.wafer_map[stream_id]['tel_type']
        if wafer_slot is None:  # Pull from detmapping cfg
            wafer_slot = runner.wafer_map[stream_id]['wafer_slot']
        fp_pars = optics.get_ufm_to_fp_pars(teltype, wafer_slot, runner.ufm_to_fp_file)
        rs = det_match.ResSet.from_solutions(sol_file, fp_pars=fp_pars, platform=teltype)
        rs.name = 'sol'
        return rs

    elif cfg.solution_type == 'resonator_set':
        sol_file = os.path.join(cfg.resonator_set_dir, f"{stream_id}.npy")
        rs_arr = np.load(sol_file)
        rs = det_match.ResSet.from_array(rs_arr)
        rs.name = 'sol'
        return rs

def get_detset_time(detset: str) -> float:
    """
    Gets timestamp associated with a detset. Will parse this from the detset
    name, assuming it is of the form <stream_id>_<time>_tune.
    """
    return float(detset.split('_')[-2])

def add_to_failed_cache(cache_file, detset, msg, cfg: UpdateDetMatchesConfig):
    if time.time() - get_detset_time(detset) < cfg.time_before_cache_failure:
        logger.info(f"{detset} is too recent to add to failed cache")
        return

    if os.path.exists(cache_file):
        with open(cache_file, 'r') as f:
            x = yaml.safe_load(f)
    else:
        x = {}  

    x[str(detset)] = str(msg)
    with open(cache_file, 'w') as f:
        yaml.dump(x, f)

def get_failed_detsets(cache_file):
    if not os.path.exists(cache_file):
        return []

    with open(cache_file, 'r') as f:
        x = yaml.safe_load(f)

    return list(x.keys())

def run_match_aman(runner: Runner, aman, detset, wafer_slot=None):
    stream_id = aman.det_info.stream_id[aman.det_info.detset == detset][0]

    rs0 = det_match.ResSet.from_aman(aman, stream_id)
    rs0.name = 'meas'

    rs1 = load_solution_set(runner, stream_id, wafer_slot=wafer_slot)
    rs1.name = 'sol'

    match_pars = det_match.MatchParams(**runner.cfg.match_pars)
    freq_offsets = runner.cfg.freq_offsets
    if freq_offsets is not None:
        costs, opt_freq = scan_for_freq_offset(
            rs0, rs1, freq_offsets, show_pb=runner.cfg.show_pb,
            match_pars=match_pars,
        )
        match_pars.freq_offset_mhz = opt_freq
    match = det_match.Match(rs0, rs1, match_pars=match_pars, 
                     apply_dst_pointing=runner.cfg.apply_solution_pointing)
    return match

def run_match(runner: Runner, detset: str) -> bool:
    """
    Creates match files for specified detset, along with any other unmatched
    detsets in the loaded observation. If match fails for a known reason, this
    will add it to the failed_detset_cache so that it is not re-attempted.

    Returns
    -------
    success: bool
        True if match was successful
    """
    # Find obs-id with cal info
    obs_all = set(runner.ctx.obsdb.query("type=='obs'")['obs_id'])
    cur = runner.ctx.obsfiledb.conn.execute(f"""
        select distinct obs_id from files where detset=='{detset}'
    """)
    obsids_with_cal = set(runner.detcal_db.get_entries(['dataset'])['dataset'])
    obs_dset = {r[0] for r in cur}
    obs_ids = sorted(list(
        obs_all.intersection(obs_dset).intersection(obsids_with_cal)), 
        key=lambda s:s.split('_')[1])[::-1]
    if len(obs_ids) == 0:
        add_to_failed_cache(
            runner.failed_detset_cache_path, detset, "NO_OBSID_WITH_CAL", runner.cfg
        )
        logger.error(f"Cannot find obsid for detset {detset}")
        return False

    obs_id = obs_ids[0]

    book_dir = os.path.dirname(runner.ctx.obsfiledb.get_files(obs_id)[detset][0][0])
    book_idx_file = os.path.join(book_dir, 'M_index.yaml')
    with open(book_idx_file, 'r') as f:
        book_idx = yaml.safe_load(f)

    aman = runner.ctx.get_meta(obs_id, ignore_missing=True)
    finished_detsets = set([os.path.splitext(f)[0] for f in os.listdir(runner.match_dir)])
    new_detsets = []
    for detset in np.unique(aman.det_info.detset):
        if detset not in finished_detsets:
            new_detsets.append(detset)

    logger.info(f"Loaded obs_id {obs_id}. Running matches for detsets:")
    for ds in new_detsets:
        logger.info(f"    - {ds}")

    for ds in new_detsets:

        stream_id = aman.det_info.stream_id[aman.det_info.detset == ds][0]
        # Try to get wafer slot info from book idx
        if 'wafer_slots' in book_idx:
            for ws in book_idx['wafer_slots']:  
                if ws['stream_id'] == stream_id:
                    wafer_slot = ws['wafer_slot']
                    break
            else:
                logger.error(
                    f"Could not find wafer_slot from book index for ds={detset}, "
                    f"obs_id={obs_id}"
                )
                raise Exception("Could not find wafer-slot")
        else:
            wafer_slot = None

        match = run_match_aman(runner, aman, ds, wafer_slot=wafer_slot)
        fpath = os.path.join(runner.match_dir, f"{ds}.h5")
        match.save(fpath)
        logger.info(f"Saved match to file: {fpath}")

    return True


def scan_for_freq_offset(rs0, rs1, freq_offsets, match_pars=None, show_pb=True):
    """
    Scans through a list of frequency offsets to find optimal match between two
    res-sets.

    Returns
    ----------
    costs : np.ndarray
        Costs of size ``len(freq_offsets)`` containing the matching cost
        at each frequency
    opt_freq : float
        Optimal offset-frequency for match.
    """
    if match_pars is None:
        match_pars = det_match.MatchParams()
    else:
        match_pars = deepcopy(match_pars)
    freq_offsets = np.array(freq_offsets)
    
    rs0 = deepcopy(rs0) 
    rs1 = deepcopy(rs1)

    costs = np.full_like(freq_offsets, np.nan)
    for i, offset in enumerate(tqdm(freq_offsets, disable=(not show_pb))):
        match_pars.freq_offset_mhz = offset
        match = det_match.Match(rs0, rs1, match_pars=match_pars)
        costs[i] = match.matching_cost

    imin = np.argmin(costs)
    opt_freq = freq_offsets[imin]
    logger.info(f"Found freq offset: {opt_freq}")
    return costs, opt_freq


def update_manifests(runner: Runner, detset):
    det_match_idx = os.path.join(runner.cfg.results_path, 'det_match.sqlite')
    det_match_h5 = os.path.join(runner.cfg.results_path, 'det_match.h5')
    assignment_idx = os.path.join(runner.cfg.results_path, 'assignment.sqlite')
    assignment_h5 = os.path.join(runner.cfg.results_path, 'assignment.h5')

    match_file = os.path.join(runner.cfg.results_path, f'matches/{detset}.h5')
    with h5py.File(match_file, 'r') as f:
        ra = np.array(f['merged'])

    names = list(ra.dtype.names)
    names[names.index('readout_id')] = 'dets:readout_id'
    names[names.index('det_id')] = 'dets:det_id'
    ra.dtype.names = tuple(names)
    assignment = ra[['dets:readout_id', 'dets:det_id']]

    def add_to_db(arr, db_path, h5_path, detset, write_relpath=True):
        write_dataset(core.metadata.ResultSet.from_friend(arr), h5_path, detset, overwrite=True)
        if not os.path.exists(db_path):
            scheme = core.metadata.ManifestScheme()
            scheme.add_exact_match('dets:detset')
            scheme.add_data_field('dataset')
            db = core.metadata.ManifestDb(db_path, scheme=scheme)

        if write_relpath:
            h5_path = os.path.relpath(h5_path, start=os.path.dirname(db_path))

        db = core.metadata.ManifestDb(db_path)
        if detset not in db.get_entries(['dataset'])['dataset']:
            db.add_entry({
                'dets:detset': detset,
                'dataset': detset
            }, h5_path)
        else:
            logger.warning(f"Dataset {detset} already exists in db: {db_path}")
    
    write_relpath = runner.cfg.write_relpath
    add_to_db(ra, det_match_idx, det_match_h5, detset,
              write_relpath=write_relpath)
    add_to_db(assignment, assignment_idx, assignment_h5, detset,
              write_relpath=write_relpath)


def update_manifests_all(runner):
    det_match_idx = os.path.join(runner.cfg.results_path, 'det_match.sqlite')
    assignment_idx = os.path.join(runner.cfg.results_path, 'assignment.sqlite')
    if os.path.exists(det_match_idx):
        det_match_db = ManifestDb(det_match_idx)
        det_match_detsets = det_match_db.get_entries(['dataset'])['dataset']
    else:
        det_match_detsets = []
    if os.path.exists(assignment_idx):
        assignment_db = ManifestDb(assignment_idx)
        assignment_detsets = assignment_db.get_entries(['dataset'])['dataset']
    else:
        assignment_detsets = []

    indexed_detsets = set(det_match_detsets).intersection(set(assignment_detsets))
    completed_detsets = set([os.path.splitext(f)[0] for f in os.listdir(runner.match_dir)])

    for ds in (completed_detsets - indexed_detsets):
        logger.info(f"Adding {ds} to manifests")
        update_manifests(runner, ds)

def make_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('config', type=str, help='path to config file')
    parser.add_argument('--all', action='store_true', help='run all detsets')
    return parser

def main(config_file: str, all: bool=False):
    with open(config_file, 'r') as f:
        cfg = UpdateDetMatchesConfig(**yaml.safe_load(f))

    runner = Runner(cfg)

    remaining_detsets = runner.get_remaining_detsets()
    logger.info(f"{len(remaining_detsets)} detsets to match")
    if all:
        update_manifests_all(runner)
        for detset in remaining_detsets:
            run_match(runner, detset)
            update_manifests_all(runner)
    else:
        for detset in remaining_detsets:
            success = run_match(runner, detset)
            update_manifests_all(runner)
            if success:
                break

if __name__ == '__main__':
    parser = make_parser()
    args = parser.parse_args()
    main(args.config, all=args.all)
